from atropos.utils.colorspace import encode, decode
from atropos.console import execute_cli


# If there are any unknown characters in the test sequence, round tripping will only
# work if all characters after the first unknown character are also unknown:
# encode("TNGN") == "T444", but
# decode("T444") == "TNNN".
sequences = [
    "",
    "C",
    "ACGGTC",
    "TN",
    "TN.",
    "TNN.N",
    "CCGGCAGCATTCATTACGACAACGTGGCACCGTGTTTTCTCGGTGGTA",
    "TGCAGTTGATGATCGAAGAAAACGACATCATCAGCCAGCAAGTGC",
    "CAGGGTTTGATGAGTGGCTGTGGGTGCTGGCGTATCCGGG",
]


def test_encode():
    assert encode("AA") == "A0"
    assert encode("AC") == "A1"
    assert encode("AG") == "A2"
    assert encode("AT") == "A3"
    assert encode("CA") == "C1"
    assert encode("CC") == "C0"
    assert encode("CG") == "C3"
    assert encode("CT") == "C2"
    assert encode("GA") == "G2"
    assert encode("GC") == "G3"
    assert encode("GG") == "G0"
    assert encode("GT") == "G1"
    assert encode("TA") == "T3"
    assert encode("TC") == "T2"
    assert encode("TG") == "T1"
    assert encode("TT") == "T0"
    assert encode("TN") == "T4"
    assert encode("NT") == "N4"
    assert encode("NN") == "N4"
    assert encode("ACGGTC") == "A13012"
    assert encode("TTT.N") == "T0044"
    assert encode("TTNT.N") == "T04444"


def test_decode():
    for s in sequences:
        expected = s.replace(".", "N")
        encoded = encode(s)
        assert decode(encoded) == expected
    assert decode("A.") == "AN"
    assert decode("C.") == "CN"
    assert decode("G.") == "GN"
    assert decode("T.") == "TN"


def test_qualtrim_csfastaqual(run_trimmer):
    run_trimmer(
        "-c -q 10", "solidqual.fastq", "solid.csfasta", qualfile="solid.qual",
    )


def test_e3m(input_data):
    execute_cli(
        [
            "-o",
            "/dev/null",
            "-se",
            input_data("E3M.fasta"),
            "-sq",
            input_data("E3M.qual"),
        ]
    )


def test_bwa(run_trimmer):
    run_trimmer(
        "-c -e 0.12 -a 330201030313112312 -x 552: --maq",
        "solidmaq.fastq",
        "solid.csfasta",
        qualfile="solid.qual",
    )


def test_bfast(run_trimmer):
    run_trimmer(
        "-c -e 0.12 -a 330201030313112312 -x abc: --strip-f3",
        "solidbfast.fastq",
        "solid.csfasta",
        qualfile="solid.qual",
    )


def test_trim_095(run_trimmer):
    run_trimmer(
        "-c -e 0.122 -a 330201030313112312",
        "solid.fasta",
        "solid.fasta",
    )


def test_solid(run_trimmer):
    run_trimmer(
        "-c -e 0.122 -a 330201030313112312",
        "solid.fastq",
        "solid.fastq",
    )


def test_solid_basespace_adapter(run_trimmer):
    run_trimmer(
        "-c -e 0.122 -a CGCCTTGGCCGTACAGCAG",
        "solid.fastq",
        "solid.fastq",
    )


def test_solid5p(run_trimmer):
    # this is not a real adapter, just a random string
    # in colorspace: C0302201212322332333
    run_trimmer(
        "-c -e 0.1 --trim-primer -g CCGGAGGTCAGCTCGCTATA",
        "solid5p.fasta",
        "solid5p.fasta",
    )


def test_solid5p_prefix_notrim(run_trimmer):
    run_trimmer(
        "-c -e 0.1 -g ^CCGGAGGTCAGCTCGCTATA",
        "solid5p-anchored.notrim.fasta",
        "solid5p.fasta",
    )


def test_solid5p_prefix(run_trimmer):
    run_trimmer(
        "-c -e 0.1 --trim-primer -g ^CCGGAGGTCAGCTCGCTATA",
        "solid5p-anchored.fasta",
        "solid5p.fasta",
    )


def test_solid5p_fastq(run_trimmer):
    # this is not a real adapter, just a random string
    # in colorspace: C0302201212322332333
    run_trimmer(
        "-c -e 0.1 --trim-primer -g CCGGAGGTCAGCTCGCTATA",
        "solid5p.fastq",
        "solid5p.fastq",
    )


def test_solid5p_prefix_notrim_fastq(run_trimmer):
    run_trimmer(
        "-c -e 0.1 -g ^CCGGAGGTCAGCTCGCTATA",
        "solid5p-anchored.notrim.fastq",
        "solid5p.fastq",
    )


def test_solid5p_prefix_fastq(run_trimmer):
    run_trimmer(
        "-c -e 0.1 --trim-primer -g ^CCGGAGGTCAGCTCGCTATA",
        "solid5p-anchored.fastq",
        "solid5p.fastq",
    )


def test_sra_fastq(run_trimmer):
    run_trimmer(
        "-c -e 0.1 --input-format sra-fastq -a CGCCTTGGCCGTACAGCAG",
        "sra.fastq",
        "sra.fastq",
    )


def test_no_zero_cap(run_trimmer):
    run_trimmer(
        "--no-zero-cap -c -e 0.122 -a CGCCTTGGCCGTACAGCAG",
        "solid-no-zerocap.fastq",
        "solid.fastq",
    )



from abc import ABCMeta, abstractmethod
from enum import Enum
from multiprocessing import Process
from typing import (
    Iterable,
    Optional,
    Sequence as SequenceType,
    Tuple,
    Type,
)

from atropos.commands import Pipeline, Summary
from atropos.commands.multicore import WorkerProcess
from atropos.commands.trim.modifiers import Modifiers
from atropos.commands.trim.filters import Filter, Filters
from atropos.commands.trim.writers import Formatters, Writers
from atropos.io.sequence import Sequence
from atropos.utils.collections import Summarizable


class CompressionMode(Enum):
    WORKER = "worker"
    WRITER = "writer"


class RecordHandler(Summarizable):
    """
    Base class for record handlers.
    """

    def __init__(self, modifiers: Modifiers, filters: Filters, formatters: Formatters):
        self.modifiers = modifiers
        self.filters = filters
        self.formatters = formatters

    def handle_record(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ) -> Tuple[Type[Filter], Tuple[Sequence, Optional[Sequence]]]:
        """
        Handles a pair of reads.
        """
        reads = self.modifiers.modify(read1, read2)
        dest = self.filters.filter(*reads)
        self.formatters.format(context["results"], dest, *reads)
        return dest, reads

    def summarize(self) -> dict:
        """
        Returns a summary dict.
        """
        return dict(
            trim=dict(
                modifiers=self.modifiers.summarize(),
                filters=self.filters.summarize(),
                formatters=self.formatters.summarize(),
            )
        )


class ResultHandler(metaclass=ABCMeta):
    """
    Base class for result handlers.
    """

    def start(self, worker: Optional[Process] = None):
        """
        Starts the result handler.
        """

    def finish(self, total_batches: Optional[int] = None):
        """
        Finishes the result handler.

        Args:
            total_batches: Total number of batches processed.
        """

    @abstractmethod
    def write_result(self, batch_num: int, result: dict):
        """
        Writes a batch of results to output.

        Args:
            batch_num: The batch number.
            result: The result to write.
        """


class ResultHandlerWrapper(ResultHandler):
    """
    Wraps a ResultHandler.
    """

    def __init__(self, handler: ResultHandler):
        self.handler = handler

    def start(self, worker: Optional[Process] = None):
        self.handler.start(worker)

    def write_result(self, batch_num: int, result: dict):
        self.handler.write_result(batch_num, result)

    def finish(self, total_batches: Optional[int] = None):
        self.handler.finish(total_batches=total_batches)


class WorkerResultHandler(ResultHandlerWrapper):
    """
    Wraps a ResultHandler and compresses results prior to writing.
    """

    def write_result(self, batch_num: int, result: dict):
        """
        Given a dict mapping files to lists of strings, joins the strings and
        compresses them (if necessary) and then returns the properly formatted
        result dict.
        """
        self.handler.write_result(
            batch_num, dict(self.prepare_file(*item) for item in result.items())
        )

    def prepare_file(self, path: str, strings: Iterable[str]):
        """
        Prepares data for writing.

        Returns:
            Tuple (path, data).
        """
        return path, "".join(strings)


class WriterResultHandler(ResultHandler):
    """
    ResultHandler that writes results to disk.
    """

    def __init__(
        self, writers: Writers, compressed: bool = False, use_suffix: bool = False
    ):
        """
        Args:
            writers: :class:`Writers` object.
            compressed: Whether the data is compressed.
            use_suffix: Whether to add the worker index as a file suffix. Used for
                parallel-write mode.
        """
        self.writers = writers
        self.compressed = compressed
        self.use_suffix = use_suffix

    def start(self, worker: Optional[WorkerProcess] = None):
        if self.use_suffix:
            if worker is None:
                raise ValueError("worker must not be None")

            self.writers.suffix = f".{worker.index}"

    def write_result(self, batch_num: int, result: dict):
        self.writers.write_result(result, self.compressed)

    def finish(self, total_batches: Optional[int] = None):
        self.writers.close()


class TrimPipeline(Pipeline, metaclass=ABCMeta):
    """
    Base trimming pipeline.

    Args:
        record_handler:
        result_handler:
    """

    def __init__(self, record_handler: RecordHandler, result_handler: ResultHandler):
        super().__init__()
        self.record_handler = record_handler
        self.result_handler = result_handler

    def start(self, worker: Optional[Process] = None):
        self.result_handler.start(worker)

    def add_to_context(self, context: dict):
        context["results"] = {}

    def handle_records(self, context: dict, records: SequenceType[Sequence]):
        super().handle_records(context, records)
        self.result_handler.write_result(context["index"], context["results"])

    def handle_reads(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        return self.record_handler.handle_record(context, read1, read2)

    def finish(self, summary: Summary, **kwargs):
        self.result_handler.finish()
        super().finish(summary)
        summary.update(self.record_handler.summarize())



from multiprocessing import Process, Queue
import os
from typing import Iterator, Optional, Type, cast

from loguru import logger

from atropos.commands import Batch, Command, Pipeline
from atropos.errors import AtroposError, MulticoreError
from atropos.utils import ReturnCode, run_interruptible
from atropos.utils.multicore import (
    DEFAULT_RETRY_INTERVAL,
    dequeue,
    enqueue,
    enqueue_all,
    ensure_processes,
    kill,
    wait_on,
)


class MulticorePipelineMixin:
    """
    Mixin that implements the `start`, `finish`, and `process_batch` methods of
    :class:`Pipeline`.
    """

    seen_batches = None

    def start(self, **kwargs):
        cast(Pipeline, super()).start(**kwargs)
        self.seen_batches = set()

    def process_batch(self, batch: Batch):
        self.seen_batches.add(batch[0]["index"])
        cast(Pipeline, super()).process_batch(batch)

    def finish(self, summary: dict, worker: "WorkerProcess" = None):
        cast(Pipeline, super()).finish(summary, worker=worker)
        logger.debug(
            f"{worker.name} finished; processed {len(self.seen_batches)} batches, "
            f"{sum(cast(Pipeline, self).record_counts.values())} reads"
        )


class WorkerProcess(Process):
    """
    Parent class for worker processes that execute Pipelines.
    """

    def __init__(
        self,
        index: int,
        input_queue: Queue,
        pipeline: Pipeline,
        summary_queue: Queue,
        timeout: int,
    ):
        """
        Args:
            index: A unique ID for the process.
            input_queue: Queue with batches of records to process.
            pipeline: The pipeline to execute.
            summary_queue: Queue where summary information is written.
            timeout: Time to wait upon queue full/empty.
        """
        super().__init__(name=f"Worker process {index}")
        self.index = index
        self.input_queue = input_queue
        self.pipeline = pipeline
        self.summary_queue = summary_queue
        self.timeout = timeout

    def run(self) -> None:
        logger.debug(f"{self.name} running under pid {os.getpid()}")
        summary = {}

        def iter_batches() -> Iterator:
            """
            Deques and yields batches.
            """
            while True:
                yield dequeue(
                    self.input_queue,
                    wait_message=f"{self.name} waiting on batch {{}}",
                    block_timeout=self.timeout,
                )

        def enqueue_summary():
            """
            Enqueues a summary dict.
            """
            enqueue(
                self.summary_queue,
                (
                    self.index,
                    cast(MulticorePipelineMixin, self.pipeline).seen_batches,
                    summary,
                ),
                wait_message=f"{self.name} waiting to queue summary {{}}",
                block_timeout=self.timeout,
            )

        try:
            self.pipeline.start(worker=self)

            try:
                for batch in iter_batches():
                    if batch is None:
                        break

                    logger.debug(
                        f"{self.name} processing batch {batch[0]['index']} of size "
                        f"{batch[0]['size']}"
                    )
                    self.pipeline.process_batch(batch)
            finally:
                self.pipeline.finish(summary, worker=self)

            logger.debug(f"{self.name} finished normally")
        except Exception as err:
            logger.exception(f"Unexpected error in {self.name}")
            summary["exception"] = err

        logger.debug(f"{self.name} sending summary")
        enqueue_summary()


class ParallelPipelineRunner:
    """
    Runs a pipeline in parallel.
    """

    def __init__(
        self, command: Command, pipeline: Pipeline, threads: Optional[int] = None
    ):
        """
        Args:
            command:
            pipeline: A :class:`Pipeline`.
            threads: Number of threads to use. If None, the value will be taken
                from command_runner.
        """
        self.command = command
        self.pipeline = pipeline
        self.threads = threads or command.get_option("threads")
        self.timeout = max(
            command.get_option("process_timeout", 0), DEFAULT_RETRY_INTERVAL
        )
        # Queue by which batches of reads are sent to worker processes
        self.input_queue = Queue(command.get_option("read_queue_size"))
        # Queue for processes to send summary information back to main process
        self.summary_queue = Queue(self.threads)
        self.worker_processes = None
        self.num_batches = None
        self.seen_summaries = None
        self.seen_batches = None

    def ensure_alive(self) -> None:
        """
        Called when enqueue times out.
        """
        ensure_processes(self.worker_processes)

    def after_enqueue(self) -> None:
        """
        Called after all batches are queued.
        """

    def finish(self) -> None:
        """
        Called at the end of the run.
        """

    def run(self) -> ReturnCode:
        """
        Runs the pipeline.

        Returns:
            The return code.
        """
        retcode = run_interruptible(self)
        self.terminate(retcode)
        return retcode

    def terminate(self, retcode: ReturnCode) -> None:
        """
        Terminates all processes.

        Args:
            retcode: The return code.
        """
        # notify all threads that they should stop
        logger.debug("Exiting all processes")
        for process in self.worker_processes:
            kill(process, retcode, self.timeout)

    def __call__(self) -> None:
        # Start worker processes, reserve a thread for the reader process,
        # which we will get back after it completes
        worker_args = (
            self.input_queue,
            self.pipeline,
            self.summary_queue,
            self.timeout,
        )
        self.worker_processes = launch_workers(self.threads - 1, worker_args)
        self.num_batches = enqueue_all(
            self.command.iterator(), self.input_queue, self.timeout, self.ensure_alive,
        )

        logger.debug(f"Main loop complete; saw {self.num_batches} batches")
        # Tell the worker processes no more input is coming
        enqueue_all(
            (None,) * self.threads, self.input_queue, self.timeout, self.ensure_alive
        )
        self.after_enqueue()
        # Now that the reader process is done, it essentially frees up another thread
        # to use for a worker
        self.worker_processes.extend(
            launch_workers(1, worker_args, offset=self.threads - 1)
        )

        # Wait for all summaries to be available on queue
        def summary_timeout_callback():
            """
            Ensures that workers are still alive.
            """
            try:
                ensure_processes(
                    self.worker_processes,
                    "Workers are still alive and haven't returned summaries: {}",
                    alive=False,
                )
            except:
                logger.exception("Error calling ensure_processes")

        wait_on(
            self.summary_queue.full,
            wait_message="Waiting on worker summaries {}",
            timeout=self.timeout,
            wait=True,
            timeout_callback=summary_timeout_callback,
        )

        logger.debug("Processing summary information from worker processes")
        self.seen_summaries = set()
        self.seen_batches = set()

        def summary_fail_callback():
            """
            Raises AtroposError with workers that did not report summaries.
            """
            missing_summaries = set(range(1, self.threads)) - self.seen_summaries
            raise AtroposError(
                f"Missing summaries from processes "
                f"{','.join(str(summ) for summ in missing_summaries)}"
            )

        for _ in range(1, self.threads + 1):
            batch = dequeue(self.summary_queue, fail_callback=summary_fail_callback)
            worker_index, worker_batches, worker_summary = batch
            if worker_summary is None:
                raise MulticoreError(f"Worker process {worker_index} died unexpectedly")
            elif (
                "exception" in worker_summary
                and worker_summary["exception"] is not None
            ):
                raise AtroposError(
                    f"Worker process {worker_index} died unexpectedly",
                    worker_summary["exception"],
                )
            else:
                logger.debug(f"Processing summary for worker {worker_index}")

            self.seen_summaries.add(worker_index)
            self.seen_batches |= worker_batches
            self.command.merge_summary(worker_summary)

        # Check if any batches were missed
        if self.num_batches > 0:
            missing_batches = set(range(1, self.num_batches + 1)) - self.seen_batches
            if len(missing_batches) > 0:
                raise AtroposError(
                    f"Workers did not process batches "
                    f"{','.join(str(batch) for batch in missing_batches)}"
                )

        self.finish()


def launch_workers(
    num_workers: int,
    args=(),
    offset: int = 0,
    worker_class: Type[WorkerProcess] = WorkerProcess
):
    """
    Launches `num_workers` workers. Each worker is initialized with an incremental
    index starting with `offset`, followed by `args`.
    """
    logger.info(f"Starting {num_workers} worker processes")
    # create workers
    workers = [worker_class(i + offset, *args) for i in range(num_workers)]
    # start workers
    for worker in workers:
        worker.start()
    return workers



# Import cythonized functions, defaulting to pure python implementations.
try:
    from atropos.commands.trim._qualtrim import quality_trim_index, twocolor_trim_index
except:
    from loguru import logger
    from typing import Tuple

    from atropos.io.sequence import Sequence
    from atropos.utils.ngs import qual2int

    logger.debug("Import failed for cythonized qualtrim functions")

    def quality_trim_index(
        qualities: str,
        cutoff_front: int,
        cutoff_back: int,
        base: int = 33
    ) -> Tuple[int, int]:
        """
        Finds the position at which to trim a low-quality end from a nucleotide
        sequence.

        Qualities are assumed to be ASCII-encoded as chr(qual + base).

        The algorithm is the same as the one used by BWA within the function
        'bwa_trim_read':
        - Subtract the cutoff value from all qualities.
        - Compute partial sums from all indices to the end of the sequence.
        - Trim sequence at the index at which the sum is minimal.
        """
        start = 0
        stop = max_i = len(qualities)

        # find trim position for 5' end
        score = 0
        max_qual = 0

        for idx in range(max_i):
            qual = qual2int(qualities[idx], base)
            score += cutoff_front - (qual - base)

            if score < 0:
                break

            if score > max_qual:
                max_qual = score
                start = idx + 1

        # same for 3' end
        max_qual = 0
        score = 0

        for idx in reversed(range(max_i)):
            qual = qual2int(qualities[idx], base)
            score += cutoff_back - (qual - base)

            if score < 0:
                break

            if score > max_qual:
                max_qual = score
                stop = idx

        if start >= stop:
            start, stop = 0, 0

        return start, stop

    def twocolor_trim_index(sequence: Sequence, cutoff: int, base: int = 33) -> int:
        """
        Variant of the above quality trimming routine that works on two-color data,
        in which bases are encoded with two colors. 'No color' (a dark cycle) usually
        means that a 'G' was sequenced, but that also occurs when sequencing falls
        off the end of the fragment. The read then contains a run of high-quality G
        bases in the end.

        This routine works as the one above, but counts qualities belonging to
        'G' bases as being equal to cutoff - 1.
        """
        bases = sequence.sequence
        qualities = sequence.qualities
        score = 0
        max_qual = 0
        max_i = len(qualities)

        for idx in reversed(range(max_i)):
            qual = qual2int(qualities[idx], base)

            if bases[idx] == "G":
                qual = cutoff - 1

            score += cutoff - qual

            if score < 0:
                break

            if score > max_qual:
                max_qual = score
                max_i = idx

        return max_i



from pathlib import Path
import traceback
from typing import List, Optional, Sequence, Tuple, Union

import pytest

from atropos.commands.trim.console import TrimCommandConsole

from .utils import assert_files_equal


class Datadir:
    def __init__(self, request, subdir: str):
        basedir = Path(str(request.fspath.dirpath())) / "data"

        self._datadirs = []

        def add(d: Path):
            s = d / subdir
            if s.exists() and s.is_dir():
                self._datadirs.append(s)

        testdir = basedir / request.module.__name__

        if request.cls is not None:
            clsdir = testdir / request.cls.__name__
            add(clsdir / request.function.__name__)
            add(clsdir)
        else:
            add(testdir / request.function.__name__)

        add(testdir)
        add(basedir)

    def __call__(self, name: str, **kwargs):
        name = name.format(**kwargs)

        for datadir in self._datadirs:
            datadir_path = datadir / name
            if datadir_path.exists():
                return datadir_path

        raise KeyError(
            f"File '{name} not found in any of the following datadirs: "
            f"{self._datadirs}"
        )

    def url(self, name: str, **kwargs):
        return f"file:{self(name, **kwargs)}"


@pytest.fixture(scope="function")
def input_data(request):
    return Datadir(request, "input")


@pytest.fixture(scope="function")
def expected_data(request):
    return Datadir(request, "expected")


@pytest.fixture(scope="function")
def run_trimmer(tmp_path: Path, capsys, input_data, expected_data):
    def _run_atropos_trim(
        params: Union[str, list],
        expected1=None,
        inpath1=None,
        inpath2=None,
        qualfile=None,
        interleaved_input=False,
        expected2=None,
        interleaved_output=False,
        aligner: str = "adapter",
        sra_accn=None,
        stdout=False,
        expected_multi: Optional[Sequence[str]] = None,
        expected_other: Optional[Sequence[str]] = None,
        assert_output_equal: bool = True,
        error_on_rc: bool = True,
    ) -> Tuple[Path, List[Path], dict]:
        if isinstance(params, str):
            param_list = params.split()
        else:
            param_list = list(params)

        to_compare = []

        output = tmp_path / ("tmp1-" + expected1)
        output2 = None
        if expected2:
            output2 = tmp_path / ("tmp2-" + expected2)

        if expected_multi:
            if stdout:
                raise ValueError("'stdout' and 'expected_multi' are mutually exclusive")

            for name in expected_multi:
                multi_name = expected1.format(name=name, aligner=aligner)
                multi_path = tmp_path / ("tmp1-" + multi_name)
                to_compare.append(
                    (expected_data(multi_name), multi_path)
                )
        elif assert_output_equal:
            to_compare.append((expected_data(expected1, aligner=aligner), output))

            if output2:
                to_compare.append(
                    (expected_data(expected2, aligner=aligner), output2)
                )

        if expected_other:
            for n, other in enumerate(expected_other, 3):
                i = param_list.index(other)
                other_path = tmp_path / (f"tmp{n}-" + other)
                param_list[i] = other_path
                to_compare.append((expected_data(other, aligner=aligner), other_path))

        infiles = [
            input_data(infile, aligner=aligner) if isinstance(infile, str) else infile
            for infile in (inpath1, inpath2)
        ]

        param_list.extend(("--aligner", aligner))

        if sra_accn:
            param_list.extend(("--accession", sra_accn))
        elif interleaved_input:
            param_list.extend(("-l", infiles[0]))
        elif inpath2:
            param_list.extend(("-pe1", infiles[0]))
            param_list.extend(("-pe2", infiles[1]))
        else:
            param_list.extend(("-se", infiles[0]))
            if qualfile:
                param_list.extend(("-sq", input_data(qualfile, aligner=aligner)))

        if stdout:
            # Output is going to stdout, so we need to redirect it to the temp file
            result = TrimCommandConsole.execute(
                tuple(str(p) for p in param_list)
            )
            captured = capsys.readouterr()
            with open(output, "wt") as out:
                out.write(captured.out)
        else:
            if interleaved_output:
                param_list.extend(("-L", output))
            else:
                param_list.extend(("-o", output))  # TODO not parallelizable
                if output2:
                    param_list.extend(("-p", output2))

            result = TrimCommandConsole.execute(
                tuple(str(p) for p in param_list)
            )

        retcode, summary = result

        assert summary is not None
        assert isinstance(summary, dict)

        if "exception" in summary and summary["exception"] is not None:
            assert retcode != 0
            err = summary["exception"]
            traceback.print_exception(*err["details"])
            if error_on_rc:
                raise AssertionError(
                    f"Return code {retcode} != 0; error: {err['message']}"
                )
        elif error_on_rc:
            assert retcode == 0

        for expected_path, actual_path in to_compare:
            assert actual_path.exists()
            assert_files_equal(expected_path, actual_path)

        return tmp_path, infiles, summary

    return _run_atropos_trim



from atropos.commands.trim.pipeline import (
    RecordHandler, WorkerResultHandler, WriterResultHandler,
)
from atropos.commands import Command, PairedEndPipelineMixin, SingleEndPipelineMixin
from atropos.commands.trim.pipeline import TrimPipeline
from atropos.commands.trim.writers import Writers
from atropos.utils import ReturnCode, run_interruptible


class SingleEndSerialTrimPipeline(SingleEndPipelineMixin, TrimPipeline):
    pass


class PairedEndSerialTrimPipeline(PairedEndPipelineMixin, TrimPipeline):
    pass


def run_serial(
    command: Command,
    paired: bool,
    record_handler: RecordHandler,
    writers: Writers,
) -> ReturnCode:
    """

    Args:
        command:
        paired:
        record_handler:
        writers:

    Returns:
        ReturnCode
    """
    if paired:
        pipeline_class = PairedEndSerialTrimPipeline
    else:
        pipeline_class = SingleEndSerialTrimPipeline

    result_handler = WorkerResultHandler(WriterResultHandler(writers))

    return run_interruptible(
        pipeline_class(record_handler, result_handler),
        command,
        raise_on_error=True
    )



from itertools import repeat
from typing import IO, Sequence, Dict, Callable

from atropos.commands.reports import (
    DefaultReportGenerator, BaseReportWriter, ReportWriter
)
from atropos.commands.legacy_reports import Printer, TitlePrinter
from atropos.utils import classproperty
from atropos.utils.argparse import Namespace


class ErrorTextReportWriter(BaseReportWriter):
    @classproperty
    def extensions(self) -> Sequence[str]:
        return "txt",

    def __init__(self, options: Namespace):
        super().__init__(options)

    def serialize(self, summary: dict, stream: IO) -> None:
        names = summary['input']['input_names'] or repeat(None)
        estimates = summary['errorrate']['estimate']
        _print = Printer(stream)
        _print_title = TitlePrinter(stream)

        input_idx = 0

        for input_idx, (estimate, details, name) in enumerate(
            zip(estimates, summary['errorrate']['details'], names), 1
        ):
            self._generate_estimator_report(
                stream, input_idx, estimate, details, _print, _print_title, name
            )

        if input_idx > 1:
            _print.newline()
            _print_title("Overall", level=0)
            total_lens = summary['errorrate']['total_len']
            overall_err = (
                sum(err * total_len for err, total_len in zip(estimates, total_lens)) /
                sum(total_lens)
            )
            print(f"Error rate: {overall_err:.2%}", file=stream)

    @staticmethod
    def _generate_estimator_report(
        stream: IO,
        input_idx: int,
        estimate: float,
        details: dict,
        _print: Printer,
        _print_title: Printer,
        input_name: str = None
    ):
        _print_indent = Printer(stream, indent='  ')
        _print.newline()
        _print_title(f"Input {input_idx}", level=0)

        if input_name:
            _print(f"File: {input_name}")

        _print(f"Error rate: {estimate:.2%}")

        if details:
            _print("Details:\n")
            per_read = details['per_read']
            per_cycle = details['per_cycle']
            _print_indent(f"StdErr: {per_read['standard error']:.2%}")
            _print_indent("Per-cycle rates:")

            for cycle in per_cycle:
                _print_indent(
                    "Cycle: {}, Error: {:.2%}, StdErr: {:.2%}".format(*cycle), indent=2
                )


class ErrorReportGenerator(DefaultReportGenerator):
    @classproperty
    def report_formats(cls) -> Dict[str, Callable[[Namespace], ReportWriter]]:
        report_formats = super().report_formats()
        report_formats["txt"] = ErrorTextReportWriter
        return report_formats



import pkg_resources


UNKNOWN_VERSION = "Unknown"


try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    __version__ = UNKNOWN_VERSION



from multiprocessing import Process, Queue
import os
from typing import Dict, Iterable, Optional

from loguru import logger
import xphyle
from xphyle.formats import CompressionFormat

from atropos.commands import Command, PairedEndPipelineMixin, SingleEndPipelineMixin
from atropos.commands.multicore import (
    MulticorePipelineMixin, ParallelPipelineRunner, WorkerProcess
)
from atropos.commands.trim.pipeline import (
    CompressionMode,
    RecordHandler,
    ResultHandler,
    WorkerResultHandler,
    WriterResultHandler,
    TrimPipeline,
)
from atropos.commands.trim.writers import Writers
from atropos.utils import ReturnCode
from atropos.utils.multicore import (
    DEFAULT_RETRY_INTERVAL,
    Control,
    ControlSignal,
    MulticoreError,
    PendingQueue,
    enqueue,
    dequeue,
    kill,
    wait_on_process,
)


class Done(MulticoreError):
    """
    Raised when process exits normally.
    """

    pass


class Killed(MulticoreError):
    """
    Raised when process exits is killed.
    """

    pass


class WriterManager:
    """
    Manager for a writer process and control variable.
    """

    def __init__(
        self,
        writers: Writers,
        compression_mode: CompressionMode,
        preserve_order: bool,
        result_queue: Queue,
        timeout: int,
    ):
        compressed = (compression_mode == CompressionMode.WORKER)

        if preserve_order:
            writer_result_handler = OrderPreservingWriterResultHandler(
                writers, compressed=compressed
            )
        else:
            writer_result_handler = WriterResultHandler(
                writers, compressed=compressed
            )

        self.timeout = timeout
        # Shared variable for communicating with writer thread
        self.writer_control = Control(ControlSignal.ACTIVE)
        # writer process
        self.writer_process = ResultProcess(
            writer_result_handler, result_queue, self.writer_control, timeout
        )
        self.writer_process.start()

    @property
    def is_active(self) -> bool:
        """
        Returns True if the writer process is alive and the control value is
        ControlSignal.ACTIVE.
        """
        return self.writer_process.is_alive() and self.writer_control.check_value(
            ControlSignal.ACTIVE
        )

    def set_num_batches(self, num_batches: int):
        """
        Sets the number of batches to the control variable.
        """
        self.writer_control.set_value(num_batches)

    def wait(self):
        """
        Waits for the writer process to terminate.
        """
        wait_on_process(self.writer_process, self.timeout)

    def terminate(self, retcode: ReturnCode):
        """
        Forces the writer process to terminate.
        """
        kill(self.writer_process, retcode, self.timeout)


class ParallelTrimPipelineRunner(ParallelPipelineRunner):
    """
    ParallelPipelineRunner for a TrimPipeline.
    """

    def __init__(
        self,
        command: Command,
        pipeline: TrimPipeline,
        threads: int,
        writer_manager: Optional[WriterManager] = None,
    ):
        super().__init__(command, pipeline, threads)
        self.writer_manager = writer_manager

    def ensure_alive(self):
        super().ensure_alive()
        if self.writer_manager and not self.writer_manager.is_active:
            raise MulticoreError("Writer process exited")

    def after_enqueue(self):
        # Tell the writer thread the max number of batches to expect
        if self.writer_manager:
            self.writer_manager.set_num_batches(self.num_batches)

    def finish(self):
        if self.writer_manager:
            # Wait for writer to complete
            self.writer_manager.wait()

    def terminate(self, retcode: ReturnCode):
        super().terminate(retcode)
        if self.writer_manager:
            self.writer_manager.terminate(retcode)


class QueueResultHandler(ResultHandler):
    """
    ResultHandler that writes results to the output queue.
    """

    def __init__(self, queue: Queue):
        self.queue = queue
        self.message = None
        self.timeout = None

    def start(self, worker: Optional[WorkerProcess] = None):
        if worker is None:
            raise RuntimeError("worker must not be None")

        self.message = f"{worker.name} waiting to queue result {{}}"
        self.timeout = worker.timeout

    def write_result(self, batch_num: int, result: dict):
        enqueue(
            self.queue,
            (batch_num, result),
            wait_message=self.message,
            timeout=self.timeout,
        )


class CompressingWorkerResultHandler(WorkerResultHandler):
    """
    Wraps a ResultHandler and compresses results prior to writing.
    """

    def __init__(self, *args, compression_format: Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.compression_format = compression_format
        self.file_compressors: Optional[Dict[str, CompressionFormat]] = None

    def start(self, worker: Optional[WorkerProcess] = None):
        super().start(worker)
        self.file_compressors = {}

    def prepare_file(self, path: str, strings: Iterable[str]):
        compressor = self.get_compressor(path)

        if compressor:
            return (
                (path, "wb"),
                compressor.compress(b"".join(s.encode() for s in strings)),
            )
        else:
            return (path, "wt"), "".join(strings)

    def get_compressor(self, filename: str) -> CompressionFormat:
        """
        Returns the file compressor based on the file extension.
        """
        if filename not in self.file_compressors:
            self.file_compressors[filename] = xphyle.get_compressor(
                self.compression_format or filename
            )

        return self.file_compressors[filename]


class OrderPreservingWriterResultHandler(WriterResultHandler):
    """
    Writer thread that is less time/memory efficient, but is guaranteed to preserve
    the original order of records.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pending = None
        self.cur_batch = None

    def start(self, worker: Optional[WorkerProcess] = None):
        super().start(worker)
        self.pending = PendingQueue()
        self.cur_batch = 1

    def write_result(self, batch_num: int, result: dict):
        if batch_num == self.cur_batch:
            self.writers.write_result(result, self.compressed)
            self.cur_batch += 1
            self.consume_pending()
        else:
            self.pending.push(batch_num, result)

    def finish(self, total_batches: Optional[int] = None):
        if total_batches is not None:
            self.consume_pending()

            if self.cur_batch != total_batches + 1:
                raise MulticoreError(
                    f"OrderPreservingWriterResultHandler finishing "
                    f"without having seen {total_batches + 1 - self.cur_batch} of "
                    f"{total_batches} batches"
                )

        super().finish(total_batches=total_batches)

    def consume_pending(self):
        """
        Consumes any remaining items in the queue.
        """
        while (not self.pending.empty) and (
            self.cur_batch == self.pending.min_priority
        ):
            self.writers.write_result(self.pending.pop(), self.compressed)
            self.cur_batch += 1


class ResultProcess(Process):
    """
    Thread that accepts results from the worker threads and process them using a
    ResultHandler. Each batch is expected to be (batch_num, path, records),
    where path is the destination file and records is a string. Not guaranteed to
    preserve the original order of sequence records.
    """

    def __init__(
        self,
        result_handler: ResultHandler,
        queue: Queue,
        control: Control,
        timeout: int = 60,
    ):
        """
        Args:
            result_handler: A ResultHandler object.
            queue: Input queue.
            control: A shared value for communcation with the main process.
            timeout: Seconds to wait for next batch before complaining.
        """
        super().__init__(name="Result process")
        self.result_handler = result_handler
        self.queue = queue
        self.control = control
        self.timeout = timeout
        self.seen_batches = set()
        self.num_batches = None

    def run(self):
        logger.debug(f"Writer process {self.name} running under pid {os.getpid()}")

        def fail_callback():
            """
            Raises Done if the expected number of batches has been seen.
            """
            if self.num_batches is None and self.control.check_value_positive():
                self.num_batches = self.control.get_value()

            if (
                self.num_batches is not None
                and len(self.seen_batches) >= self.num_batches
            ):
                raise Done()

        def timeout_callback():
            """
            Logs an error with the missing batches.
            """
            if self.num_batches is not None:
                missing = set(range(1, self.num_batches + 1)) - self.seen_batches
                logger.error(
                    f"Result thread still missing batches"
                    f"{','.join(str(i) for i in missing)} of {self.num_batches}"
                )

        def iter_batches():
            """
            Dequeues and yields batches.
            """
            while True:
                yield dequeue(
                    self.queue,
                    wait_message="Result process waiting on result {}",
                    timeout=self.timeout,
                    fail_callback=fail_callback,
                    timeout_callback=timeout_callback,
                )

        try:
            self.result_handler.start(self)
            for batch_num, result in iter_batches():
                self.seen_batches.add(batch_num)
                self.result_handler.write_result(batch_num, result)
        except Done:
            logger.debug("Writer process exiting normally")
        except Killed:
            logger.debug("Writer process exited early")
        except:
            logger.exception("Unexpected error in writer process")
            self.control.set_value(ControlSignal.ERROR)
        finally:
            num_batches = self.control.get_value(lock=True)
            self.result_handler.finish(num_batches if num_batches > 0 else None)


class SingleEndMulticoreTrimPipeline(
    MulticorePipelineMixin, SingleEndPipelineMixin, TrimPipeline
):
    pass


class PairedEndMulticoreTrimPipeline(
    MulticorePipelineMixin, PairedEndPipelineMixin, TrimPipeline
):
    pass


def run_multicore(
    trim_command: Command,
    paired: bool,
    record_handler: RecordHandler,
    writers: Writers,
    threads: int,
    result_queue_size: int,
    writer_process: bool,
    preserve_order: bool,
    compression_mode: CompressionMode,
    compression_format: str,
    process_timeout: int,
) -> ReturnCode:
    """
    Parallel implementation of run_atropos. Works as follows:

    1. Main thread creates N worker processes (where N is the number of
    threads to be allocated) and (optionally) one writer process.
    2. Main thread loads batches of reads (or read pairs) from input file(s)
    and adds them to a queue (the input queue).
    3. Worker processes take batches from the input queue, process them as
    Atropos normally does, and either add the results to the result queue
    (if using a writer process) or write the results to disk. Each result is
    a dict mapping output file names to strings, where each string is a
    concatenation of reads (with appropriate line endings) to be written.
    A parameter also controls whether data compression is done by the workers or
    the writer.
    4. If using a writer process, it takes results from the result queue and
    writes each string to its corresponding file.
    5. When the main process finishes loading reads from the input file(s),
    it sends a signal to the worker processes that they should complete when
    the input queue is empty. It also singals the writer process how many
    total batches to expect, and the writer process exits after it has processed
    that many batches.
    6. When a worker process completes, it adds a summary of its activity to
    the summary queue.
    7. The main process reads summaries from the summary queue and merges
    them to create the complete summary, which is used to generate the report.

    There are several possible points of failure:

    1. The main process may exit due to an unexpected error, or becuase the
    user forces it to exit (Ctrl-C). In this case, an attempt is made to
    cancel all processes before exiting.
    2. A worker or writer process may exit due to an unknown error. To
    handle this, the main process checks that each process is alive whenver
    it times out writing to the input queue, and again when waiting for
    worker summaries. If a process has died, the program exits with an error
    since some data might have gotten lost.
    3. More commonly, process will time out blocking on reading from or
    writing to a queue. Size limits are used (optionally) for the input and
    result queues to prevent using lots of memory. When few threads are
    allocated, it is most likely that the main and writer processes will
    block, whereas with many threads allocated the workers are most likely
    to block. Also, e.g. in a cluster environment, I/O latency may cause a
    "backup" resulting in frequent blocking of the main and workers
    processes. Finally, also e.g. in a cluster environment, processes may
    suspended for periods of time. Use of a hard timeout period, after which
    processes are forced to exit, is thus undesirable. Instead, parameters
    are provided for the user to tune the batch size and max queue sizes to
    their particular environment. Additionally, a "soft" timeout is used,
    after which log messages are escallated from DEBUG to ERROR level. The
    user can then make the decision of whether or not to kill the program.

    Args:
        threads:
        process_timeout:
        compression_mode:
        result_queue_size:
        writer_process:
        compression_format:
        writers: Writers object
        preserve_order:
        paired: Whether the input is paired-end.
        record_handler: RecordHandler object.
        trim_command:

    Returns:
        The return code.
    """
    timeout = max(process_timeout, DEFAULT_RETRY_INTERVAL)
    logger.debug(
        f"Starting atropos in multicore mode with threads={threads}, "
        f"timeout={timeout}"
    )

    if threads < 2:
        raise ValueError("'threads' must be >= 2")

    # Reserve a thread for the writer process if it will be doing the compression
    # and if one is available.
    if compression_mode == CompressionMode.WRITER and threads > 2:
        threads -= 1

    # Queue by which results are sent from the worker processes to the writer
    # process
    result_queue = Queue(result_queue_size)
    writer_manager = None

    if writer_process:
        if compression_mode == CompressionMode.WRITER:
            worker_result_handler = WorkerResultHandler(
                QueueResultHandler(result_queue)
            )
        else:
            worker_result_handler = CompressingWorkerResultHandler(
                QueueResultHandler(result_queue),
                compression_format=compression_format
            )

        writer_manager = WriterManager(
            writers,
            compression_mode,
            preserve_order,
            result_queue,
            timeout,
        )
    else:
        worker_result_handler = WorkerResultHandler(
            WriterResultHandler(writers, use_suffix=True)
        )

    if paired:
        pipeline_class = PairedEndMulticoreTrimPipeline
    else:
        pipeline_class = SingleEndMulticoreTrimPipeline

    pipeline = pipeline_class(record_handler, worker_result_handler)

    runner = ParallelTrimPipelineRunner(
        trim_command, pipeline, threads, writer_manager
    )

    return runner.run()



from abc import ABCMeta, abstractmethod
import copy
from multiprocessing import cpu_count
from pathlib import Path
import platform
import re
import textwrap
from typing import Dict, Sequence as SequenceType, Tuple, Type, Union, cast

from loguru import logger
from xphyle.paths import STDOUT, STDERR

from atropos import __version__
from atropos.commands import BaseCommand
from atropos.io import InputRead, SequenceFileType
from atropos.utils import (
    LOGGING_CONFIG, ReturnCode, classproperty
)
from atropos.utils.argparse import (
    AtroposArgumentParser,
    EnumNameAction,
    EnumValueAction,
    Namespace,
    ParagraphHelpFormatter,
    int_or_str,
    positive,
    probability,
    readable_file,
    writable_file,
)
from atropos.utils.ngs import ALPHABETS
from atropos.utils.paths import splitext_compressed
from atropos.utils.progress import ProgressBarType


class CommandConsole(metaclass=ABCMeta):
    @classproperty
    @abstractmethod
    def name(cls) -> str:
        """
        The command name.
        """

    @classmethod
    @abstractmethod
    def get_help(
        cls, fmt: str = "* {name}: {description}", wrap: int = 80, indent: int = 2
    ) -> str:
        """
        Returns a string to include in the command help.
        """

    @classmethod
    @abstractmethod
    def execute(cls, args: SequenceType[str] = ()) -> Tuple[ReturnCode, dict]:
        """
        Parse command line arguments, execute the command, and generate summary reports.

        Returns:
            A tuple (return_code, summary)
        """


class BaseCommandConsole(metaclass=ABCMeta):
    """
    Base class for Atropos commands.
    """

    @classproperty
    @abstractmethod
    def name(cls) -> str:
        """
        The command name.
        """

    @classproperty
    @abstractmethod
    def description(cls) -> str:
        """
        The command description.
        """

    @classproperty
    def usage(cls) -> str:
        """
        Command usage string.
        """
        return "atropos {command} [options]"

    @classproperty
    def preamble(cls) -> str:
        """
        Help message preamble.
        """
        return "Atropos version {version}"

    @classproperty
    def details(cls) -> str:
        """
        Help message details.
        """
        return ""

    @classmethod
    def get_description(cls, **kwargs) -> str:
        description = "{}\n\n{}\n\n{}".format(
            *(part.strip() for part in (cls.preamble, cls.description, cls.details))
        )
        return description.format(**kwargs)

    @classmethod
    def get_help(
        cls, fmt: str = "* {name}: {description}", wrap: int = 80, indent: int = 2
    ) -> str:
        """
        Returns a string to include in the command help.

        Args:
            fmt: The help string format.
            wrap: The number of characters at which to wrap lines.
            indent: The number of spaces to indent each line.
        """
        helpstr = fmt.format(name=cls.name, description=cls.description.strip())

        if wrap:
            return "\n".join(
                textwrap.wrap(
                    re.sub(r"\s+", " ", helpstr), wrap, subsequent_indent=" " * indent
                )
            )
        else:
            return helpstr

    @classmethod
    def execute(cls, args: SequenceType[str] = ()) -> Tuple[ReturnCode, dict]:
        """
        Parse command line arguments, execute the command, and generate summary reports.

        Returns:
            The tuple (return_code, summary)
        """
        options = cls._parse_args(args)
        command = cls._create_command(options)
        retcode, summary = command.run()
        if retcode is ReturnCode.SUCCESS and options.report_file:
            logger.debug(f"Writing report to {options.report_file}")
            cls.generate_reports(summary, options)
        else:
            logger.debug("Not generating report file")
        return retcode, summary

    @classmethod
    def _parse_args(cls, args: SequenceType[str]) -> Namespace:
        """
        Parses command line arguments.

        Args:
            args: command-line arguments

        Returns:
            A `Namespace`
        """
        parser = cls._create_argument_parser()
        cls._add_arguments(parser)

        options = parser.parse_args(args)
        options.orig_args = copy.copy(args)

        cls._setup_logging(options)
        cls._validate_options(options, parser)

        return options

    @classmethod
    def _create_command(cls, options: Namespace) -> BaseCommand:
        if issubclass(cls, BaseCommand):
            return cast(Type[BaseCommand], cls)(options)
        else:
            raise NotImplementedError()

    @classmethod
    def _create_argument_parser(cls) -> AtroposArgumentParser:
        """
        Creates an `ArgumentParser` to parse command line arguments to this command.

        Returns:
            An `ArgumentParser`
        """
        format_args = dict(name=cls.name, version=__version__)
        return AtroposArgumentParser(
            prog=f"atropos {cls.name}",
            usage=cls.usage.format(**format_args),
            description=cls.get_description(**format_args),
            formatter_class=ParagraphHelpFormatter,
        )

    @classmethod
    @abstractmethod
    def _add_arguments(cls, parser: AtroposArgumentParser) -> None:
        """
        Adds arguments to the ArgumentParser.

        Args:
            parser:
        """

    @classmethod
    def _validate_options(
        cls, options: Namespace, parser: AtroposArgumentParser
    ) -> None:
        """
        Validates command-line options.

        Args:
            options:
        """

    @classmethod
    def _setup_logging(cls, options: Namespace) -> None:
        """
        Setup logging and print an introductory message.

        Logging setup is only done if there are not already any handlers (can
        happen when this function is being called externally such as from unit
        tests).

        Args:
            options: Command-line options
        """
        if not LOGGING_CONFIG.was_setup:
            level = options.log_level or ("ERROR" if options.quiet else "INFO")

            if options.log_file is None:
                # Due to backwards compatibility, log messages are sent to
                # standard output instead of standard error if the -o option is
                # used.
                if options.output is STDERR:
                    log_path = STDOUT
                else:
                    log_path = STDERR
            else:
                log_path = options.log_file

            LOGGING_CONFIG.setup(log_path, level)

        logger.info(
            f"This is Atropos {__version__} with Python {platform.python_version()}"
        )

    @classmethod
    @abstractmethod
    def generate_reports(cls, summary: dict, options: Namespace) -> None:
        """
        Generates summary report(s) for this command.

        Args:
            summary:
            options:
        """


def add_common_options(parser: AtroposArgumentParser) -> None:
    """
    Adds common arguments to the parser.
    """
    parser.set_defaults(
        orig_args=None,
        paired=False,
        default_outfile=STDOUT,
        report_file=None,
        report_formats=None,
        batch_size=1000,
        counter_magnitude="M",
        ngstream_reader=None,
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Print debugging information. (no)",
    )
    parser.add_argument(
        "--progress",
        nargs="?",
        action=EnumNameAction,
        const=ProgressBarType,
        default=ProgressBarType.NONE,
        help="Show progress. An optional argument determines what type of progress "
        "meter to use: bar = progress bar; msg = status message; any = show any type "
        "of progress meter; none = do not show a progress meter. (none)"
    )
    parser.add_argument(
        "--quiet",
        action="store_true",
        default=False,
        help="Print only error messages. (no)",
    )
    parser.add_argument(
        "--log-level",
        choices=tuple(logger._core.levels.keys()),  # TODO: accessing protected member
        default=None,
        help="Logging level. (ERROR when --quiet else INFO)",
    )
    parser.add_argument(
        "--log-file",
        type=writable_file,
        default=None,
        metavar="FILE",
        help="File to write logging info. (stdout)",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Show version information and exit.",
    )
    group = parser.add_group("Input")
    group.add_argument(
        "-pe1",
        "--input1",
        type=readable_file,
        default=None,
        metavar="FILE1",
        help="The first input file.",
    )
    group.add_argument(
        "-pe2",
        "--input2",
        type=readable_file,
        default=None,
        metavar="FILE2",
        help="The second input file.",
    )
    group.add_argument(
        "-l",
        "--interleaved-input",
        type=readable_file,
        default=None,
        metavar="FILE",
        help="Interleaved input file.",
    )
    group.add_argument(
        "-se",
        "--single-input",
        type=readable_file,
        default=None,
        metavar="FILE",
        help="A single-end read file.",
    )
    group.add_argument(
        "--single-input-read",
        action=EnumValueAction,
        const=(InputRead, (InputRead.READ1, InputRead.READ2)),
        dest="input_read",
        default=None,
        help="When treating an interleaved FASTQ or paired-end SAM/BAM "
        "file as single-end, this option specifies which of the two "
        "reads to process. (both reads used)",
    )
    group.add_argument(
        "-sq",
        "--single-quals",
        type=readable_file,
        default=None,
        metavar="FILE",
        help="A single-end qual file.",
    )
    group.add_argument(
        "--umi-input",
        type=readable_file,
        default=None,
        metavar="FILE",
        help="Read UMIs from a separate input file."
    )
    group.add_argument(
        "--accession",
        default=None,
        metavar="ACCN",
        help="Accession to stream using a supported protocol. Should be of the form "
        "<protocol>:<accession>, e.g. 'sra:SRR000066'. If no protocol is specified, "
        "it assumed to be an SRA accession."
    )
    group.add_argument(
        "--query",
        default=None,
        metavar="URL",
        help="Query URL for a supported protocol. Should be of the form "
        "'<protocol>+http://...'. If no protocol is specified, it is assumed to be "
        "htsget."
    )
    group.add_argument(
        "-f",
        "--input-format",
        action=EnumNameAction,
        const=SequenceFileType,
        default=None,
        help="Input file format. Ignored when reading csfasta/qual files. "
        "(auto-detect from file name extension)",
    )
    group.add_argument(
        "-Q",
        "--quality-base",
        type=positive(),
        default=33,
        help="Assume that quality values in FASTQ are encoded as "
        "ascii(quality + QUALITY_BASE). This needs to be set to 64 "
        "for some old Illumina FASTQ files. (33)",
    )
    group.add_argument(
        "-c",
        "--colorspace",
        action="store_true",
        default=False,
        help="Enable colorspace mode: Also trim the color that is adjacent "
        "to the found adapter. (no)",
    )
    group.add_argument(
        "--max-reads",
        type=int_or_str,
        default=None,
        metavar="N",
        help="Maximum number of reads/pairs to process (no max)",
    )
    group.add_argument(
        "--subsample",
        type=probability,
        default=None,
        metavar="PROB",
        help="Subsample a fraction of reads. (no)",
    )
    group.add_argument(
        "--subsample-seed",
        type=int,
        default=None,
        metavar="SEED",
        help="The seed to use for the pseudorandom number generator. Using "
        "the same seed will result in the same subsampling of reads.",
    )
    group.add_argument(
        "--batch-size",
        type=int_or_str,
        metavar="SIZE",
        help="Number of records to process in each batch. (1000)",
    )
    group.add_argument(
        "-D",
        "--sample-id",
        default=None,
        metavar="ID",
        help="Optional sample ID. Added to the summary output.",
    )
    group.add_argument(
        "--alphabet",
        default=None,
        metavar="NAME",
        choices=tuple(ALPHABETS.keys()),
        help="Specify a sequence alphabet to use for validating inputs. "
        "Currently, only 'dna' is supported. (no validation)",
    )


def validate_common_options(options: Namespace, parser: AtroposArgumentParser) -> None:
    """
    Validates arguments to common options.
    """
    # Find out which 'mode' we need to use.
    # TODO: unit tests for SRA streaming
    if options.accession or options.query:
        if options.input_format is None:
            options.input_format = SequenceFileType.FASTQ
        elif options.input_format not in {
            SequenceFileType.FASTQ,
            SequenceFileType.SRA_FASTQ,
            SequenceFileType.SAM,
            SequenceFileType.BAM,
        }:
            raise ValueError(
                f"Invalid file format for accession/query: {options.input_format}"
            )

        if options.accession:
            accession = options.accession

            if ":" in accession:
                protocol, accession = accession.split(":", 1)
            else:
                protocol = "sra"

            logger.debug(
                f"Opening reader for {protocol} Accession {options.accession}"
            )
        else:
            accession = options.query

            if "+" in accession:
                protocol, accession = accession.split("+", 1)
            else:
                protocol = "htsget"

            logger.debug(
                f"Opening reader for {protocol} query {options.accession}"
            )

        try:
            import ngstream

            options.ngstream_reader = ngstream.open(
                accession, protocol, batch_size=options.batch_size or 1000
            )
            options.ngstream_reader.start()
            options.paired = options.ngstream_reader.paired
        except Exception:
            logger.exception(
                f"Error while fetching accession {options.sra_accession} from SRA"
            )
            parser.error(f"Unable to read from accession {options.sra_accession}")
    elif options.single_input:
        if options.input1 or options.input2 or options.interleaved_input:
            parser.error("Cannot use -se together with -pe1, -pe2, or -l")

        options.paired = False
        options.input1 = options.single_input
        options.input2 = options.single_quals
    elif options.interleaved_input and options.input_read:
        options.input1 = options.interleaved_input
        options.paired = False
    else:
        if not options.interleaved_input and not (options.input1 and options.input2):
            parser.error(
                "Both '-pe1' and '-pe2' are required for paired-end "
                "trimming. If this is an interleaved file, use '-l' "
                "instead."
            )

        options.paired = True

    if options.input_read is None:
        options.input_read = InputRead.PAIRED if options.paired else InputRead.SINGLE

    # Set sample ID from the input file name(s)
    if options.sample_id is None:
        if options.ngstream_reader:
            options.sample_id = options.ngstream_reader.name
        else:
            fname = Path(options.input1 or options.interleaved_input).name
            name = splitext_compressed(fname)[0]

            if options.input2:
                name2 = splitext_compressed(Path(options.input2).name)[0]
                if name != name2:
                    for i in range(min(len(name), len(name2))):
                        if name[i] != name2[i]:
                            name = name[:i]
                            break

            if name.endswith("."):
                name = name[:-1]

            options.sample_id = name

    if options.quiet:
        options.progress = ProgressBarType.NONE
    elif options.progress != ProgressBarType.NONE and options.output == STDERR:
        logger.warning("Progress bar may corrupt output written to STDERR")

    if options.report_file in (STDOUT, STDERR) and options.quiet:
        logger.warning("Quiet mode - report will not be written to stdout")
        options.report_file = None


def configure_threads(options, parser) -> int:
    """
    Determines the number of threads to use from the command-line options. Updates
    the value in options, and returns the number of threads.
    """
    if options.debug:
        parser.error("Cannot use debug mode with multiple threads")

    threads = options.threads

    if threads <= 0:
        threads = cpu_count()
    elif threads == 1:
        parser.error("--threads must be >= 2")

    options.threads = threads

    return threads


def parse_metrics_args(args_str) -> Dict[str, Union[bool, str]]:
    """
    Parses the optional value to the '--stat' option.
    """
    args = {}

    for arg in args_str.split(";"):
        arg_parts = arg.split("=")
        if len(arg_parts) == 1:
            args[arg_parts[0]] = True
        else:
            args[arg_parts[0]] = arg_parts[1]

    return args



from collections import Sequence as SequenceCollection
from enum import Enum
from pathlib import Path
import sys
from typing import Dict, Optional, Tuple, Type, Union, cast

from loguru import logger

from atropos.io import SequenceFileType
from xphyle import STDOUT, STDERR

from atropos.adapters import AdapterParser, AdapterType
from atropos.commands import BaseCommand, Summary
from atropos.commands.metrics import (
    MetricsMode,
    SingleEndReadMetrics,
    PairedEndReadMetrics,
)
from atropos.commands.trim.filters import (
    Filter,
    FilterFactory,
    Filters,
    MergedReadFilter,
    NContentFilter,
    NoFilter,
    TooLongReadFilter,
    TooShortReadFilter,
    TrimmedFilter,
    UntrimmedFilter,
)
from atropos.commands.trim.modifiers import (
    AdapterCutter,
    DoubleEncoder,
    InsertAdapterCutter,
    LengthTagModifier,
    MergeOverlapping,
    MinCutter,
    NEndTrimmer,
    TwoColorQualityTrimmer,
    UmiTrimmer,
    SyncUmi,
    AddUmi,
    NonDirectionalBisulfiteTrimmer,
    OverwriteRead,
    PairedEndModifiers,
    PrefixSuffixAdder,
    PrimerTrimmer,
    QualityTrimmer,
    RRBSTrimmer,
    SingleEndModifiers,
    SuffixRemover,
    SwiftBisulfiteTrimmer,
    UnconditionalCutter,
    ZeroCapper,
    Modifiers,
)
from atropos.commands.trim.pipeline import (
    CompressionMode,
    RecordHandler,
    ResultHandler,
    TrimPipeline,
    WorkerResultHandler,
    WriterResultHandler,
)
from atropos.commands.trim.writers import (
    Formatters,
    InfoFormatter,
    RestFormatter,
    WildcardFormatter,
    Writers,
)
from atropos.io.sequence import Sequence
from atropos.io.readers import SAMReader
from atropos.utils import ReturnCode, classproperty
from atropos.utils.argparse import Namespace
from atropos.utils.collections import Summarizable
from atropos.utils.statistics import RandomMatchProbability


class MetricsRecordHandlerWrapper(Summarizable):
    """
    Wrapper around a record handler that collects read metrics before and/or after
    trimming.
    """

    def __init__(
        self,
        record_handler: RecordHandler,
        paired: bool,
        metrics_args: Dict[MetricsMode, dict],
        **kwargs,
    ):
        """
        Args:
            record_handler:
            paired: Whether reads are paired-end.
            metrics_args: Sequence; when to collect metrics ('pre', 'post')
            kwargs:
                mode: Collection mode; pre=only collect pre-trim metrics; post=only
                collect post-trim metrics; both=collect both pre- and post-trim metrics.
        """
        self.record_handler = record_handler
        self.read_metrics_class = (
            PairedEndReadMetrics if paired else SingleEndReadMetrics
        )

        if MetricsMode.PRE in metrics_args:
            self.pre = {}
            self.pre_kwargs = kwargs.copy()
            self.pre_kwargs.update(metrics_args[MetricsMode.PRE])
        else:
            self.pre = None

        if MetricsMode.POST in metrics_args:
            self.post = {}
            self.post_kwargs = kwargs.copy()
            self.post_kwargs.update(metrics_args[MetricsMode.POST])
        else:
            self.post = None

    def handle_record(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ) -> Tuple[Type[Filter], Tuple[Sequence, Optional[Sequence]]]:
        """
        Handles a pair of reads.

        Args:
            context:
            read1:
            read2:
        """
        if self.pre is not None:
            self.collect(self.pre, context["source"], read1, read2, **self.pre_kwargs)
        dest, reads = self.record_handler.handle_record(context, read1, read2)
        if self.post is not None:
            if dest not in self.post:
                self.post[dest] = {}
            self.collect(self.post[dest], context["source"], *reads, **self.post_kwargs)
        return dest, reads

    def collect(
        self,
        metrics: dict,
        source: Union[str, Tuple[str, str]],
        read1: Sequence,
        read2: Optional[Sequence] = None,
        **kwargs,
    ):
        """
        Collects metrics on a pair of reads.

        Args:
            metrics: The :class:`ReadStatistics` object.
            source: The source file(s).
            read1
            read2: The reads.
            kwargs:
        """
        if source not in metrics:
            metrics[source] = self.read_metrics_class(**kwargs)

        metrics[source].collect(read1, read2)

    def summarize(self) -> dict:
        """
        Returns a summary dict.
        """
        summary = self.record_handler.summarize()

        if self.pre is not None:
            summary["pre"] = dict(
                (source, metrics.summarize()) for source, metrics in self.pre.items()
            )

        if self.post is not None:
            summary["post"] = {}
            for dest, metrics_dict in self.post.items():
                summary["post"][dest.name] = dict(
                    (source, metrics.summarize())
                    for source, metrics in metrics_dict.items()
                )

        return summary


class TrimSummary(Summary):
    """
    Summary that adds aggregate values for record and bp metrics.
    """

    def _post_process_other(self, dict_val: dict, key, value) -> None:
        """
        For trim metrics, any value with a name that starts with 'records_' will have
        'fraction_<var>' computed as value / total_records, and any value with a name
        that starts with 'bp_' will have 'fraction_<var>' and 'total_<var>' computed.
        We also replace any `Const`s with their values.
        """
        if self.has_exception:
            return

        def frac(val: int, _total: int):
            """
            Computes fraction of total.
            """
            return (val / _total) if val and _total != 0 else 0

        if isinstance(key, str):
            if key.startswith("records_"):
                frac_key = f"fraction_{key}"
                total_records = self["total_record_count"]

                if isinstance(value, SequenceCollection):
                    dict_val[frac_key] = [frac(val, total_records) for val in value]
                    total = sum(val for val in value if val)
                    dict_val[f"total_{key}"] = total
                else:
                    dict_val[frac_key] = frac(value, total_records)
            elif key.startswith("bp_"):
                frac_key = f"fraction_{key}"
                sum_total_bp = self["sum_total_bp_count"]

                if isinstance(value, SequenceCollection):
                    dict_val[frac_key] = [
                        frac(val, bps)
                        for val, bps in zip(value, self["total_bp_counts"])
                    ]
                    total = sum(val for val in value if val)
                    dict_val[f"total_{key}"] = total
                    dict_val[f"fraction_total_{key}"] = frac(
                        total, sum_total_bp
                    )
                else:
                    dict_val[frac_key] = frac(value, sum_total_bp)


class AlignerType(Enum):
    ADAPTER = "adapter"
    INSERT = "insert"


class TrimOperation(Enum):
    OVERWRITE = "W"
    ADAPTER_TRIM = "A"
    CUT = "C"
    TWOCOLOR_TRIM = "G"
    QUALITY_TRIM = "Q"


class PairFilter(Enum):
    ANY = "any"
    BOTH = "both"


class TrimCommand(BaseCommand):
    @classproperty
    def name(cls) -> str:
        return "trim"

    @classmethod
    def _create_summary(cls) -> dict:
        return TrimSummary()

    def __call__(self) -> ReturnCode:
        options = self.options
        match_probability = RandomMatchProbability()

        # Create Adapters
        adapters1 = []
        adapters2 = []
        has_adapters1 = bool(options.adapters or options.anywhere or options.front)
        has_adapters2 = bool(options.adapters2 or options.anywhere2 or options.front2)

        if has_adapters1 or has_adapters2:
            adapter_cache = self._load_known_adapters()

            parser_args = dict(
                colorspace=options.colorspace,
                max_error_rate=options.error_rate,
                min_overlap=options.overlap,
                read_wildcards=options.match_read_wildcards,
                adapter_wildcards=options.match_adapter_wildcards,
                indels=options.indels,
                indel_cost=options.indel_cost,
                cache=adapter_cache,
                gc_content=options.gc_content,
                match_probability=match_probability,
                alphabet=options.alphabet,
            )
            if options.adapter_max_rmp:
                parser_args["max_rmp"] = options.adapter_max_rmp

            adapter_parser = AdapterParser(**parser_args)

            if has_adapters1:
                adapters1 = adapter_parser.parse_multi(
                    options.adapters, options.anywhere, options.front
                )

            if has_adapters2:
                adapters2 = adapter_parser.parse_multi(
                    options.adapters2, options.anywhere2, options.front2
                )

            if options.cache_adapters:
                adapter_cache.save()

        # Create Modifiers
        # TODO: can this be replaced with an argparse required group?
        if (
            not adapters1
            and not adapters2
            and not options.quality_cutoff
            and options.twocolor_trim is None
            and options.cut == []
            and options.cut2 == []
            and options.cut_min == []
            and options.cut_min2 == []
            and (options.minimum_length is None or options.minimum_length <= 0)
            and options.maximum_length == sys.maxsize
            and not options.trim_n
            and not self.get_option("has_qualfile")
            and options.max_n is None
            and (not options.paired or options.overwrite_low_quality is None)
        ):
            raise ValueError("You need to provide at least one adapter sequence.")

        if options.aligner == AlignerType.INSERT and any(
            not a or len(a) != 1 or a[0].where != AdapterType.BACK
            for a in (adapters1, adapters2)
        ):
            raise ValueError(
                "Insert aligner requires a single 3' adapter for each read"
            )

        if options.debug:
            for adapter in adapters1 + adapters2:
                adapter.enable_debug()

        if options.paired:
            modifiers = PairedEndModifiers(options.paired)
        else:
            modifiers = SingleEndModifiers()

        trim_umis = options.read1_umi or options.read2_umi

        if trim_umis:
            modifiers.add_modifier_pair(
                UmiTrimmer,
                dict(number_of_bases=options.read1_umi),
                dict(number_of_bases=options.read2_umi),
            )

        if trim_umis or options.umi_input:
            if options.paired:
                modifiers.add_modifier(SyncUmi, delim=options.umi_delim)
            else:
                modifiers.add_modifier(AddUmi, delim=options.umi_delim)

        for oper in options.op_order:
            if oper == TrimOperation.OVERWRITE and options.overwrite_low_quality:
                lowq, highq, window = options.overwrite_low_quality
                modifiers.add_modifier(
                    OverwriteRead,
                    worse_read_min_quality=lowq,
                    better_read_min_quality=highq,
                    window_size=window,
                    base=options.quality_base,
                )
            elif oper == TrimOperation.ADAPTER_TRIM and (adapters1 or adapters2):
                # TODO: generalize this using some kind of factory class
                if options.aligner == AlignerType.INSERT:
                    # Use different base probabilities if we're trimming
                    # bisulfite data.
                    # TODO: this doesn't seem to help things, so commenting it
                    #  out for now
                    # if options.bisulfite:
                    #   base_probs = dict(match_prob=0.33, mismatch_prob=0.67)
                    # else:
                    #   base_probs = dict(match_prob=0.25, mismatch_prob=0.75)
                    modifiers.add_modifier(
                        InsertAdapterCutter,
                        adapter1=adapters1[0],
                        adapter2=adapters2[0],
                        action=options.action,
                        mismatch_action=options.correct_mismatches,
                        max_insert_mismatch_frac=options.insert_match_error_rate,
                        max_adapter_mismatch_frac=
                        options.insert_match_adapter_error_rate,
                        match_probability=match_probability,
                        insert_max_rmp=options.insert_max_rmp,
                        read_wildcards=options.match_read_wildcards,
                        adapter_wildcards=options.match_adapter_wildcards,
                    )
                elif options.aligner == AlignerType.ADAPTER:
                    a1_args = (
                        dict(
                            adapters=adapters1,
                            times=options.times,
                            action=options.action,
                        )
                        if adapters1
                        else None
                    )
                    a2_args = (
                        dict(
                            adapters=adapters2,
                            times=options.times,
                            action=options.action,
                        )
                        if adapters2
                        else None
                    )
                    modifiers.add_modifier_pair(AdapterCutter, a1_args, a2_args)
            elif oper == TrimOperation.CUT and (options.cut or options.cut2):
                modifiers.add_modifier_pair(
                    UnconditionalCutter,
                    dict(lengths=options.cut),
                    dict(lengths=options.cut2),
                )
            elif (
                oper == TrimOperation.TWOCOLOR_TRIM
                and (options.twocolor_trim is not None)
            ):
                modifiers.add_modifier(
                    TwoColorQualityTrimmer,
                    cutoff=options.twocolor_trim,
                    base=options.quality_base,
                )
            elif oper == TrimOperation.QUALITY_TRIM and options.quality_cutoff:
                modifiers.add_modifier(
                    QualityTrimmer,
                    cutoff_front=options.quality_cutoff[0],
                    cutoff_back=options.quality_cutoff[1],
                    base=options.quality_base,
                )

        if options.bisulfite:
            if isinstance(options.bisulfite, str):
                if "non-directional" in options.bisulfite:
                    modifiers.add_modifier(
                        NonDirectionalBisulfiteTrimmer,
                        rrbs=options.bisulfite == "non-directional-rrbs",
                    )
                elif options.bisulfite == "rrbs":
                    modifiers.add_modifier(RRBSTrimmer)
                elif options.bisulfite in ("epignome", "truseq"):
                    # Trimming leads to worse results
                    # modifiers.add_modifier(TruSeqBisulfiteTrimmer)
                    pass
                elif options.bisulfite == "swift":
                    modifiers.add_modifier(SwiftBisulfiteTrimmer)
            else:
                if options.bisulfite[0]:
                    modifiers.add_modifier(MinCutter, read=1, **(options.bisulfite[0]))
                if len(options.bisulfite) > 1 and options.bisulfite[1]:
                    modifiers.add_modifier(MinCutter, read=2, **(options.bisulfite[1]))

        if options.trim_n:
            modifiers.add_modifier(NEndTrimmer)

        if options.cut_min or options.cut_min2:
            modifiers.add_modifier_pair(
                MinCutter, dict(lengths=options.cut_min), dict(lengths=options.cut_min2)
            )

        if options.length_tag:
            modifiers.add_modifier(LengthTagModifier, length_tag=options.length_tag)

        if options.strip_suffix:
            modifiers.add_modifier(SuffixRemover, suffixes=options.strip_suffix)

        if options.prefix or options.suffix:
            modifiers.add_modifier(
                PrefixSuffixAdder, prefix=options.prefix, suffix=options.suffix
            )

        if options.double_encode:
            modifiers.add_modifier(DoubleEncoder)

        delivers_qualities = self.get_option("delivers_qualities")

        if options.zero_cap and delivers_qualities:
            modifiers.add_modifier(ZeroCapper, quality_base=options.quality_base)

        if options.trim_primer:
            modifiers.add_modifier(PrimerTrimmer)

        if options.merge_overlapping:
            modifiers.add_modifier(
                MergeOverlapping,
                min_overlap=options.merge_min_overlap,
                error_rate=options.merge_error_rate,
                mismatch_action=options.correct_mismatches,
            )

        # Create Formatters

        interleaved = options.use_interleaved_output

        if interleaved:
            output1 = options.interleaved_output
            output2 = None
        else:
            output1 = options.output
            output2 = options.paired_output

        if options.output_format is None:
            if delivers_qualities:
                options.output_format = SequenceFileType.FASTQ
            else:
                options.output_format = SequenceFileType.FASTA

        seq_formatter_args = dict(
            file_format=options.output_format,
            qualities=delivers_qualities,
            colorspace=options.colorspace,
            interleaved=interleaved,
        )
        if (
            options.output_format == SequenceFileType.SAM
            and isinstance(self.reader, SAMReader)
        ):
            seq_formatter_args["bam_header"] = cast(SAMReader, self.reader).header

        formatters = Formatters(output1, seq_formatter_args)

        # Create filters
        min_affected = 2 if options.pair_filter == PairFilter.BOTH else 1
        filters = Filters(FilterFactory(options.paired, min_affected))
        force_create = []

        if options.merge_overlapping:
            filters.add_filter(MergedReadFilter)
            if options.merged_output:
                formatters.add_seq_formatter(MergedReadFilter, options.merged_output)

        if options.minimum_length is not None and options.minimum_length > 0:
            filters.add_filter(TooShortReadFilter, options.minimum_length)
            if options.too_short_output:
                formatters.add_seq_formatter(
                    TooShortReadFilter,
                    options.too_short_output,
                    options.too_short_paired_output,
                )

        if options.maximum_length < sys.maxsize:
            filters.add_filter(TooLongReadFilter, options.maximum_length)
            if options.too_long_output is not None:
                formatters.add_seq_formatter(
                    TooLongReadFilter,
                    options.too_long_output,
                    options.too_long_paired_output,
                )

        if options.max_n is not None:
            filters.add_filter(NContentFilter, options.max_n)

        if options.discard_trimmed:
            filters.add_filter(TrimmedFilter)

        if not formatters.multiplexed:
            if output1 is not None:
                formatters.add_seq_formatter(NoFilter, output1, output2)
                if output1 != STDOUT and options.writer_process:
                    force_create.append(output1)
                    if output2 is not None:
                        force_create.append(output2)
            elif not (options.discard_trimmed and options.untrimmed_output):
                formatters.add_seq_formatter(NoFilter, options.default_outfile)
                if (
                    options.default_outfile not in (STDOUT, STDERR) and
                    options.writer_process
                ):
                    force_create.append(options.default_outfile)

        if options.discard_untrimmed or options.untrimmed_output:
            filters.add_filter(UntrimmedFilter)

        if not options.discard_untrimmed:
            if formatters.multiplexed:
                untrimmed = (
                    options.untrimmed_output or
                    Path(str(output1).format(name="unknown"))
                )
                formatters.add_seq_formatter(UntrimmedFilter, untrimmed)
                formatters.add_seq_formatter(NoFilter, untrimmed)
            elif options.untrimmed_output:
                formatters.add_seq_formatter(
                    UntrimmedFilter,
                    options.untrimmed_output,
                    options.untrimmed_paired_output,
                )

        if options.rest_file:
            formatters.add_detail_formatter(RestFormatter(options.rest_file))

        if options.info_file:
            formatters.add_detail_formatter(InfoFormatter(options.info_file))

        if options.wildcard_file:
            formatters.add_detail_formatter(WildcardFormatter(options.wildcard_file))

        # Create writers
        writers = Writers(force_create, options.compression_format)

        # Create record handler
        record_handler = RecordHandler(modifiers, filters, formatters)

        if options.metrics:
            record_handler = MetricsRecordHandlerWrapper(
                record_handler,
                options.paired,
                options.metrics,
                qualities=delivers_qualities,
                quality_base=options.quality_base,
            )

        num_adapters = sum(len(a) for a in modifiers.get_adapters())
        trim_mode = {
            False: "single-end",
            "first": "paired-end legacy",
            "both": "paired-end",
        }[options.paired]
        logger.info(
            f"Trimming {num_adapters} adapter{'s' if num_adapters > 1 else ''} "
            f"with at most {options.error_rate * 100:.1f} errors in {trim_mode} mode..."
        )

        if options.paired == "first" and (
            len(record_handler.modifiers.get_modifiers(read=2)) > 0
            or options.quality_cutoff
        ):
            logger.warning(
                """Requested read modifications are applied only to the first read 
                since backwards compatibility mode is enabled. To modify both reads, 
                also use any of the -A/-B/-G/-U options. Use a dummy adapter sequence 
                when necessary: -A XXX"""
            )

        if options.threads is None:
            self.summary.update(mode="serial", threads=1)

            from atropos.commands.trim.serial import run_serial

            return run_serial(
                self,
                options.paired,
                record_handler,
                writers,
            )
        else:
            self.summary.update(mode="parallel", threads=options.threads)

            from atropos.commands.trim.multicore import run_multicore

            compression_mode = (
                self.get_option("compression_mode") or
                choose_compression_mode(self.options)
            )

            return run_multicore(
                self,
                options.paired,
                record_handler,
                writers,
                self.get_option("threads"),
                self.get_option("result_queue_size"),
                self.get_option("writer_process"),
                self.get_option("preserve_order"),
                compression_mode,
                self.get_option("compression_format"),
                self.get_option("process_timeout"),
            )


def choose_compression_mode(options: Namespace) -> CompressionMode:
    if (
        options.output is None or
        options.output in (STDOUT, STDERR) or (
            options.writer_process and
            2 < options.threads < 8 and
            options.can_use_system_compression
        )
    ):
        # We must use a writer process to write to stdout/stderr. Otherwise, our
        # tests show that with 8 or more threads, worker compression is more efficient.
        return CompressionMode.WRITER
    else:
        return CompressionMode.WORKER



"""
Classes for writing and filtering of processed reads.

A Filter is a callable that has the read as its only argument. If it is called,
it returns True if the read should be filtered (discarded), and False if not.

To be used, a filter needs to be wrapped in one of the Wrapper classes. To determine
what happens to a read, a list of Wrappers with different filters is created and
each Wrapper is called in turn until one returns True. The main program will
determine whether and where to write the read(s) based on whether it was rejected
by a filter, and which one.
"""
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from typing import Callable, Optional, Type

from atropos.io.sequence import Sequence
from atropos.utils import classproperty
from atropos.utils.collections import Summarizable


# Constants used when returning from a Filter’s __call__ method to improve
# readability (it is unintuitive that "return True" means "discard the read").
DISCARD = True
KEEP = False


class Filter(metaclass=ABCMeta):
    """
    Marker type for Filters.
    """

    @classproperty
    def name(cls) -> str:
        return cls.__name__

    @abstractmethod
    def __call__(self, read: Sequence) -> bool:
        pass


class MergedReadFilter(Filter):
    """
    Returns True if the read was merged.
    """

    def __call__(self, read: Sequence) -> bool:
        return read.merged


class TooShortReadFilter(Filter):
    """
    Returns True if the read sequence is shorter than `minimum_length`.
    """

    @classproperty
    def name(cls) -> str:
        return "too_short"

    def __init__(self, minimum_length: int):
        self.minimum_length = minimum_length

    def __call__(self, read: Sequence) -> bool:
        return len(read) < self.minimum_length


class TooLongReadFilter(Filter):
    """
    Returns True if the read sequence is longer than `maximum_length`.
    """

    @classproperty
    def name(cls) -> str:
        return "too_long"

    def __init__(self, maximum_length: int):
        self.maximum_length = maximum_length

    def __call__(self, read: Sequence) -> bool:
        return len(read) > self.maximum_length


class NContentFilter(Filter):
    """
    Discards a reads that has a number of 'N's over a given threshold. It handles
    both raw counts of Ns as well as proportions. Note, for raw counts, it is a
    greater than comparison, so a cutoff of '1' will keep reads with a single N in it.
    """

    @classproperty
    def name(cls) -> str:
        return "too_many_n"

    def __init__(self, count: int):
        """
        Args:
            count: If it is below 1.0, it will be considered a proportion, and above
            and equal to 1 will be considered as discarding reads with a number of
            N's greater than this cutoff.
        """
        if count < 0:
            raise ValueError(f"Count {count} is < 0")

        self.is_proportion = count < 1.0
        self.cutoff = count

    def __call__(self, read: Sequence) -> bool:
        """
        Returns True when the read should be discarded.
        """
        n_count = read.sequence.lower().count("n")
        if not self.is_proportion:
            return n_count > self.cutoff
        elif len(read) == 0:
            return KEEP
        else:
            return n_count / len(read) > self.cutoff


class UntrimmedFilter(Filter):
    """
    Returns True if read is untrimmed.
    """

    def __call__(self, read: Sequence) -> bool:
        return read.match is None


class TrimmedFilter(Filter):
    """
    Returns True if read is trimmed.
    """

    def __call__(self, read: Sequence) -> bool:
        return read.match is not None


class NoFilter(Filter):
    """
    Always returns False.
    """

    @classproperty
    def name(cls) -> str:
        return "NoFilter"

    def __call__(self, read: Sequence) -> bool:
        return KEEP


class FilterWrapper(Summarizable, metaclass=ABCMeta):
    """
    Base wrapper around a filter.
    """

    def __init__(self, f: Filter):
        self._filtered = 0
        self._wrapped = f

    def __call__(self, read1, read2=None):
        """
        Calls the filter function.

        Returns:
            DISCARD if the filter function returns True, else KEEP
        """
        if self._filter(read1, read2):
            self._filtered += 1
            return DISCARD

        return KEEP

    @abstractmethod
    def _filter(self, read1: Sequence, read2: Optional[Sequence] = None) -> bool:
        """
        Calls the filter function.
        """

    @property
    def name(self) -> str:
        """
        The filter name.
        """
        return self._wrapped.name

    def summarize(self) -> dict:
        """
        Returns a summary dict.
        """
        return dict(records_filtered=self._filtered)


class SingleWrapper(FilterWrapper):
    """This is for single-end reads and for paired-end reads, using the 'legacy'
    filtering mode (backwards compatibility). That is, if the first read matches
    the filtering criteria, the pair is discarded. The second read is not
    inspected.
    """

    def _filter(self, read1: Sequence, read2: Optional[Sequence] = None) -> bool:
        return self._wrapped(read1)


class PairedWrapper(FilterWrapper):
    """
    This is for paired-end reads, using the 'new-style' filtering where both reads
    are inspected. That is, the entire pair is discarded if at least 1 or 2 of the
    reads match the filtering criteria.
    """

    def __init__(self, f, min_affected: int = 1):
        """
        Args:
            min_affected: values 1 and 2 are allowed.
                1 means: the pair is discarded if any read matches
                2 means: the pair is discarded if both reads match
        """
        if min_affected not in (1, 2):
            raise ValueError("min_affected must be 1 or 2")

        super().__init__(f)
        self.min_affected = min_affected

    def _filter(self, read1: Sequence, read2: Optional[Sequence] = None) -> bool:
        failures = 0

        if self._wrapped(read1):
            failures += 1

        if (self.min_affected - failures == 1) and (
            read2 is None or self._wrapped(read2)
        ):
            failures += 1

        return failures >= self.min_affected


class FilterFactory:
    """
    Factory that creates filters and wraps them in the appropriate (single-end or
    paired-end) wrapper.
    """

    def __init__(self, paired: bool, min_affected: int):
        self.paired = paired
        self.min_affected = min_affected

    def __call__(self, filter_type: Callable[..., Filter], *args, **kwargs):
        """
        Creates a wrapped filter of the specified type.
        """
        fltr = filter_type(*args, **kwargs)
        if self.paired == "both":
            return PairedWrapper(fltr, self.min_affected)
        else:
            return SingleWrapper(fltr)


class Filters(Summarizable):
    """
    Manages multiple filters.

    Args:
        filter_factory: The :class:`FilterFactory` to use for creating new filters.
    """

    def __init__(self, filter_factory: FilterFactory):
        self.filters = OrderedDict()
        self.filter_factory = filter_factory

    def add_filter(self, filter_type: Type[Filter], *args, **kwargs):
        """
        Add a filter of the specified type.
        """
        self.filters[filter_type] = self.filter_factory(filter_type, *args, **kwargs)

    def filter(self, read1: Sequence, read2: Optional[Sequence] = None) -> Type[Filter]:
        """
        Executes filters in the order they were added until one returns True.

        Args:
            read1: The first read to filter.
            read2: The second read to filter.

        Returns:
            The type of the first filter that returned True, or :class:NoFilter
            if none of the filters returned True.
        """
        for filter_type, fltr in self.filters.items():
            if fltr(read1, read2):
                return filter_type
        else:
            return NoFilter

    def __contains__(self, filter_type: Type[Filter]) -> bool:
        return filter_type in self.filters

    def __getitem__(self, filter_type: Type[Filter]) -> Filter:
        return self.filters[filter_type]

    def summarize(self) -> dict:
        """
        Returns a summary dict.
        """
        return dict((f.name, f.summarize()) for f in self.filters.values())



import math
from typing import Dict, Sequence, Optional

from atropos.errors import NotInAlphabetError


class Alphabet:
    def __init__(
        self, valid_characters: Sequence[str], default_character: Optional[str] = None
    ):
        if not isinstance(valid_characters, set):
            valid_characters = set(valid_characters)
        if default_character not in valid_characters:
            valid_characters.add(default_character)
        self.valid_characters = valid_characters
        self.default_character = default_character

    def __contains__(self, character: str) -> bool:
        return character in self.valid_characters

    def validate(self, character: str) -> None:
        """
        Raises NotInAlphabetError if the character is not in the alphabet.
        """
        if character not in self:
            raise NotInAlphabetError(character)

    def validate_string(self, string: str) -> None:
        """
        Raises NotInAlphabetError if any character in 'string' is not in the alphabet.
        """
        for character in string:
            self.validate(character)

    def resolve(self, character: str) -> str:
        """
        Returns 'character' if it's in the alphabet, otherwise the alphabet's default
        character.
        """
        if character in self.valid_characters:
            return character

        else:
            return self.default_character

    def resolve_string(self, string: str) -> str:
        """
        Returns a new string with any non-alphabet characters replaced with the
        default character.
        """
        return "".join(self.resolve(c) for c in string)


ALPHABETS = dict(
    dna=Alphabet("ACGT", "N"),
    iso=None,
    colorspace=Alphabet("0123", None)
)


def build_iso_nucleotide_table() -> Dict[str, str]:
    """
    Generates a dict mapping ISO nucleotide characters to their complements, in both
    upper and lower case.

    TODO: the nucleotide table should be implemented as an alphabet.
    """
    nuc = {
        "A": "T",
        "C": "G",
        "R": "Y",
        "S": "S",
        "W": "W",
        "K": "M",
        "B": "V",
        "D": "H",
        "N": "N",
    }
    for base, comp in tuple(nuc.items()):
        nuc[comp] = base
        nuc[base.lower()] = comp.lower()
        nuc[comp.lower()] = base.lower()
    return nuc


BASE_COMPLEMENTS = build_iso_nucleotide_table()
IUPAC_BASES = frozenset(("X",) + tuple(BASE_COMPLEMENTS.keys()))
"""Valid IUPAC bases, plus 'X'"""
GC_BASES = frozenset("CGRYSKMBDHVN")
"""IUPAC bases that include C or G."""
LOG2 = math.log(2)


def complement(seq: str) -> str:
    """
    Returns the complement of nucleotide sequence `seq`.
    """
    return "".join(BASE_COMPLEMENTS[base] for base in seq)


def reverse_complement(seq: str) -> str:
    """
    Returns the reverse complement of nucleotide sequence `seq`.
    """
    return "".join(BASE_COMPLEMENTS[base] for base in reversed(seq))


def sequence_complexity(seq: str) -> float:
    """
    Computes a simple measure of sequence complexity.

    Args:
        seq: The sequence to measure.

    Returns:
        Complexity, as a value [0,2], where 0 = a homopolymer and
        2 = completely random.
    """
    seq = seq.upper()
    seqlen = float(len(seq))
    term = 0
    for base in ("A", "C", "G", "T"):
        count = seq.count(base)
        if count > 0:
            frac = count / seqlen
            term += frac * math.log(frac) / LOG2
    return -term


def qual2int(qual: str, base: int = 33) -> int:
    """
    Converts a quality charater to a phred-scale int.

    Args:
        qual: The quality value.
        base: The offset of the first quality value (Old Illumina = 64,
            new Illumina = 33).

    Returns:
        An iterable of integer qualities.
    """
    return ord(qual) - base


def quals2ints(quals: Sequence[str], base: int = 33) -> Sequence[int]:
    """
    Converts an iterable of quality characters to phred-scale ints.

    Args:
        quals: The qualities.
        base: The offset of the first quality value (Old Illumina = 64,
            new Illumina = 33).

    Returns:
        A tuple of integer qualities.
    """
    return tuple(ord(q) - base for q in quals)


def qual2prob(qchar: str) -> float:
    """
    Converts a quality char to a probability.
    """
    return 10 ** (-qual2int(qchar) / 10)



import sys
from typing import cast

from loguru import logger
from xphyle import STDOUT, STDERR, FORMATS

from atropos.adapters import DEFAULT_ADAPTER_CACHE_FILE
from atropos.commands.console import (
    BaseCommandConsole,
    add_common_options,
    configure_threads,
    parse_metrics_args,
    validate_common_options,
)
from atropos.commands.legacy_reports import LegacyReportGenerator
from atropos.commands.metrics import MetricsMode
from atropos.commands.trim import (
    AlignerType, PairFilter, TrimCommand, TrimOperation, choose_compression_mode
)
from atropos.commands.trim.modifiers import MismatchAction, TrimAction
from atropos.commands.trim.pipeline import CompressionMode
from atropos.io import SequenceFileType
from atropos.utils import classproperty
from atropos.utils.argparse import (
    AtroposArgumentParser,
    EnumNameAction,
    EnumCharList,
    Delimited,
    Namespace,
    int_or_str,
    positive,
    probability,
    readable_file,
    readwritable_file,
    writable_file,
)


DEFAULT_ADAPTER_MAX_RMP = 1e-6
DEFAULT_ADAPTER_INDEL_COST = 1
DEFAULT_ADAPTER_MIN_OVERLAP = 3
DEFAULT_ADAPTER_MIN_OVERLAP_WITH_RMP = 1

DEFAULT_INSERT_MAX_RMP = 1e-6
DEFAULT_INSERT_MATCH_ERROR_RATE = 0.2
DEFAULT_INSERT_MIN_OVERLAP = 1
DEFAULT_INSERT_INDEL_COST = 3

DEFAULT_MERGE_MIN_OVERLAP = 0.9
DEFAULT_MERGE_ERROR_RATE = 0.2

DEFAULT_GC_CONTENT = 0.5
DEFAULT_ERROR_RATE = 0.1

DEFAULT_MIRNA_ADAPTERS = ["TGGAATTCTCGG"]  # illumina small RNA adapter
DEFAULT_MINRA_QUALITY_CUTOFF = [20, 20]
DEFAULT_MIRNA_MIN_LENGTH = 16
DEFAULT_MIRNA_ERROR_RATE = 0.12

DEFAULT_BATCH_SIZE = 1000
MAX_QUEUE_SIZE = 10e6
DEFAULT_PROCESS_TIMEOUT = 60
WRITER_READ_QUEUE_SIZE_MULTIPLER = 100
WORKER_READ_QUEUE_SIZE_MULTIPLIER = 500
WORKER_RESULT_QUEUE_SIZE_MULTIPLIER = 100
WRITER_RESULT_QUEUE_SIZE_MULTIPLIER = 500


class TrimCommandConsole(TrimCommand, LegacyReportGenerator, BaseCommandConsole):
    @classproperty
    def description(cls) -> str:
        return """
        Trim adapters and low-quality bases, and perform other NGS preprocessing. This
        command provides most of Atropos' functionality.
        """

    @classproperty
    def details(cls) -> str:
        return """
        Replace "ADAPTER" with the actual sequence of your 3' adapter. IUPAC wildcard
        characters are supported. The reverse complement is *not* automatically
        searched. All reads from input.fastq will be written to output.fastq with the
        adapter sequence removed. Adapter matching is error-tolerant. Multiple adapter
        sequences can be given (use further -a options), but only the best-matching
        adapter will be removed.
        
        Input may also be in FASTA, SAM, or BAM format. Compressed input and output is
        supported and auto-detected from the file name (.gz, .xz, .bz2). Use the file
        name '-' for standard input/output. Without the -o option, output is sent to
        standard output.
        """

    @classproperty
    def usage(cls) -> str:
        return """
        atropos trim -a ADAPTER [options] [-o output.fastq] -se input.fastq
        atropos trim -a ADAPT1 -A ADAPT2 [options] -o out1.fastq -p out2.fastq \
          -pe1 in1.fastq -pe2 in2.fastq
        """

    @classmethod
    def _add_arguments(cls, parser: AtroposArgumentParser) -> None:
        add_common_options(parser)
        cls._add_trim_options(parser)

    @classmethod
    def _validate_options(
        cls, options: Namespace, parser: AtroposArgumentParser
    ) -> None:
        validate_common_options(options, parser)
        cls._validate_trim_options(options, parser)

    @classmethod
    def _add_trim_options(cls, parser: AtroposArgumentParser):
        parser.set_defaults(
            zero_cap=None,
            action=TrimAction.TRIM,
            batch_size=None,
            known_adapter=None,
            use_interleaved_output=False,
            can_use_system_compression=False,
        )

        group = parser.add_group(
            "Adapters",
            title="Finding adapters",
            description="Parameters -a, -g, -b specify adapters to be removed from "
            "each read (or from the first read in a pair if data is "
            "paired). If specified multiple times, only the best matching "
            "adapter is trimmed (but see the --times option). When the "
            "special notation 'file:FILE' is used, adapter sequences are "
            "read from the given FASTA file. When the --adapter-file "
            "option is used, adapters can be specified by name rather "
            "than sequence.",
        )

        group.add_argument(
            "-a",
            "--adapter",
            action="append",
            default=[],
            metavar="ADAPTER",
            dest="adapters",
            help="Sequence of an adapter ligated to the 3' end (paired data: "
            "of the first read). The adapter and subsequent bases are "
            "trimmed. If a '$' character is appended ('anchoring'), the "
            "adapter is only found if it is a suffix of the read. (none)",
        )
        group.add_argument(
            "-g",
            "--front",
            action="append",
            default=[],
            metavar="ADAPTER",
            help="Sequence of an adapter ligated to the 5' end (paired data: "
            "of the first read). The adapter and any preceding bases are "
            "trimmed. Partial matches at the 5' end are allowed. If a "
            "'^' character is prepended ('anchoring'), the adapter is "
            "only found if it is a prefix of the read. (none)",
        )
        group.add_argument(
            "-b",
            "--anywhere",
            action="append",
            default=[],
            metavar="ADAPTER",
            help="Sequence of an adapter that may be ligated to the 5' or 3' "
            "end (paired data: of the first read). Both types of matches "
            "as described under -a und -g are allowed. If the first base "
            "of the read is part of the match, the behavior is as with "
            "-g, otherwise as with -a. This option is mostly for rescuing "
            "failed library preparations - do not use if you know which "
            "end your adapter was ligated to! (none)",
        )
        group.add_argument(
            "-F",
            "--known-adapters-file",
            type=readable_file,
            action="append",
            default=None,
            help="Path or URL of a FASTA file containing adapter sequences.",
        )
        group.add_argument(
            "--no-default-adapters",
            action="store_false",
            dest="default_adapters",
            default=True,
            help="Don't fetch the default adapter list (which is currently "
            "stored in GitHub).",
        )
        group.add_argument(
            "--adapter-cache-file",
            type=readwritable_file,
            default=DEFAULT_ADAPTER_CACHE_FILE,
            help="File where adapter sequences will be cached, unless "
            "--no-cache-adapters is set.",
        )
        group.add_argument(
            "--no-cache-adapters",
            action="store_false",
            dest="cache_adapters",
            default=True,
            help=f"Don't cache adapters list as '{DEFAULT_ADAPTER_CACHE_FILE}' in the "
            f"working directory.",
        )
        group.add_argument(
            "--no-trim",
            action="store_const",
            dest="action",
            const=TrimAction.NONE,
            help="Match and redirect reads to output/untrimmed-output as "
            "usual, but do not remove adapters. (no)",
        )
        group.add_argument(
            "--mask-adapter",
            action="store_const",
            dest="action",
            const=TrimAction.MASK,
            help="Mask adapters with 'N' characters instead of trimming them. (no)",
        )
        group.add_argument(
            "--gc-content",
            type=probability,
            default=DEFAULT_GC_CONTENT,
            help=f"Expected GC content of sequences. ({DEFAULT_GC_CONTENT})",
        )

        # Arguments specific to the choice of aligner

        group.add_argument(
            "--aligner",
            action=EnumNameAction,
            const=AlignerType,
            default=AlignerType.ADAPTER,
            metavar="TYPE",
            help="Which alignment algorithm to use for identifying adapters. "
            "Currently, you can choose between the semi-global alignment "
            "strategy used in Cutdapt ('adapter') or the more accurate "
            "insert-based alignment algorithm ('insert'). Note that "
            "insert-based alignment can only be used with paired-end "
            "reads containing 3' adapters. New algorithms are being "
            "implemented and the default is likely to change. (adapter)",
        )

        # Arguments for adapter match
        # TODO: all the different matching options are pretty confusing. Either
        #  explain their usage better in the docs or find a way to simplify the choices.
        group.add_argument(
            "-e",
            "--error-rate",
            type=probability,
            default=None,
            help=f"Maximum allowed error rate for adapter match (no. of errors "
            f"divided by the length of the matching region). ({DEFAULT_ERROR_RATE})",
        )
        group.add_argument(
            "--indel-cost",
            type=positive(int, True),
            default=None,
            metavar="COST",
            help=f"Integer cost of insertions and deletions during adapter "
            f"match. Substitutions always have a cost of 1. "
            f"({DEFAULT_ADAPTER_INDEL_COST})",
        )
        group.add_argument(
            "--no-indels",
            action="store_false",
            dest="indels",
            default=True,
            help="Allow only mismatches in alignments. (allow both mismatches "
            "and indels)",
        )
        group.add_argument(
            "-n",
            "--times",
            type=positive(int, False),
            default=1,
            metavar="COUNT",
            help="Remove up to COUNT adapters from each read. (1)",
        )
        group.add_argument(
            "--match-read-wildcards",
            action="store_true",
            default=False,
            help="Interpret IUPAC wildcards in reads. (no)",
        )
        group.add_argument(
            "-N",
            "--no-match-adapter-wildcards",
            action="store_false",
            dest="match_adapter_wildcards",
            default=True,
            help="Do not interpret IUPAC wildcards in adapters. (no)",
        )
        group.add_argument(
            "-O",
            "--overlap",
            type=positive(int, False),
            default=None,
            metavar="MINLENGTH",
            help=f"If the overlap between the read and the adapter is shorter "
            f"than MINLENGTH, the read is not modified. Reduces the no. "
            f"of bases trimmed due to random adapter matches. "
            f"({DEFAULT_ADAPTER_MIN_OVERLAP})",
        )
        group.add_argument(
            "--adapter-max-rmp",
            type=probability,
            default=None,
            metavar="PROB",
            help=f"If no minimum overlap (-O) is specified, then adapters are "
            f"only matched when the probabilty of observing k out of n "
            f"matching bases is <= PROB. ({DEFAULT_ADAPTER_MAX_RMP})",
        )

        # Arguments for insert match
        group.add_argument(
            "--insert-max-rmp",
            type=probability,
            default=DEFAULT_INSERT_MAX_RMP,
            metavar="PROB",
            help=f"Overlapping inserts only match when the probablity of "
            f"observing k of n matching bases is <= PROB. ({DEFAULT_INSERT_MAX_RMP})",
        )
        group.add_argument(
            "--insert-match-error-rate",
            type=probability,
            default=None,
            help=f"Maximum allowed error rate for insert match (no. of errors "
            f"divided by the length of the matching region). "
            f"({DEFAULT_INSERT_MATCH_ERROR_RATE})",
        )
        group.add_argument(
            "--insert-match-adapter-error-rate",
            type=probability,
            default=None,
            help=f"Maximum allowed error rate for matching adapters after "
            f"successful insert match (no. of errors divided by the length "
            f"of the matching region). ({DEFAULT_INSERT_MATCH_ERROR_RATE})",
        )

        # Arguments for merging and error correction
        # TODO: add RMP parameter for MergeOverlap
        group.add_argument(
            "-R",
            "--merge-overlapping",
            action="store_true",
            default=False,
            help="Merge read pairs that overlap into a single sequence. This "
            "is experimental. (no)",
        )
        group.add_argument(
            "--merge-min-overlap",
            type=positive(float, True),
            default=DEFAULT_MERGE_MIN_OVERLAP,
            help=f"The minimum overlap between reads required for merging. If this "
            f"number is (0,1.0], it specifies the minimum length as the fraction of "
            f"the length of the *shorter* read in the pair; otherwise it specifies "
            f"the minimum number of overlapping base pairs (with an absolute minimum "
            f"of 2 bp). ({DEFAULT_MERGE_MIN_OVERLAP})",
        )
        group.add_argument(
            "--merge-error-rate",
            type=probability,
            default=None,
            help=f"The maximum error rate for merging. ({DEFAULT_MERGE_ERROR_RATE})",
        )
        group.add_argument(
            "--correct-mismatches",
            action=EnumNameAction,
            const=MismatchAction,
            default=MismatchAction.NONE,
            metavar="ACTION",
            help="How to handle mismatches while aligning/merging. 'Liberal' "
            "and 'conservative' error correction both involve setting the "
            "base to the one with the best quality. They differ only when "
            "the qualities are equal -- liberal means set it to the base "
            "from the read with the overall best median base quality, "
            "while conservative means to leave it unchanged. 'N' means to "
            "set the base to N. If exactly one base is ambiguous, the "
            "non-ambiguous base is always used. (no error correction)",
        )

        group = parser.add_group("Modifications", title="Additional read modifications")
        group.add_argument(
            "--op-order",
            type=EnumCharList(TrimOperation),
            default=(
                TrimOperation.CUT,
                TrimOperation.TWOCOLOR_TRIM,
                TrimOperation.QUALITY_TRIM,
                TrimOperation.ADAPTER_TRIM,
                TrimOperation.OVERWRITE
            ),
            help="The order in which trimming operations are be applied. This "
            "is a string of the following characters: A = adapter trimming; C = "
            "cutting (unconditional); G = two-color polyG trimming; Q = quality "
            "trimming; W = overwrite poor quality reads. The default is 'WCGQA' to "
            "maintain compatibility with Cutadapt; however, this is likely to change "
            "to 'GAWCQ' in the near future. Note that UMI clipping is always performed "
            "before any other trimming operation.",
        )
        group.add_argument(
            "-u",
            "--cut",
            type=int,
            action="append",
            default=[],
            metavar="LENGTH",
            help="Remove bases from each read (first read only if paired). "
            "If LENGTH is positive, remove bases from the beginning. "
            "If LENGTH is negative, remove bases from the end. "
            "Can be used twice if LENGTHs have different signs. (no)",
        )
        group.add_argument(
            "-q",
            "--quality-cutoff",
            type=Delimited(data_type=positive(int, True), min_len=1, max_len=2),
            default=None,
            metavar="[5'CUTOFF,]3'CUTOFF",
            help="Trim low-quality bases from 5' and/or 3' ends of each read "
            "before adapter removal. Applied to both reads if data is "
            "paired. If one value is given, only the 3' end is trimmed. "
            "If two comma-separated cutoffs are given, the 5' end is "
            "trimmed with the first cutoff, the 3' end with the second. "
            "(no)",
        )
        group.add_argument(
            "-i",
            "--cut-min",
            type=int,
            action="append",
            default=[],
            metavar="LENGTH",
            help="Similar to -u, except that cutting is done AFTER adapter "
            "trimming, and only if a minimum of LENGTH bases was not "
            "already removed. (no)",
        )
        group.add_argument(
            "--twocolor-trim",
            type=positive(),
            default=None,
            metavar="3'CUTOFF",
            help="Two-color-chemistry-specific quality trimming (each read). Trims "
            "dark cycles appearing as high-quality G bases. Similar to "
            "'--quality-cutoff' but G bases are ignored. (no)",
        )
        group.add_argument(
            "--trim-n",
            action="store_true",
            default=False,
            help="Trim N's on ends of reads. (no)",
        )
        group.add_argument(
            "-x",
            "--prefix",
            default="",
            help="Add this prefix to read names. Use {name} to insert the name "
            "of the matching adapter. (no)",
        )
        group.add_argument(
            "-y",
            "--suffix",
            default="",
            help="Add this suffix to read names; can also include {name}. (no)",
        )
        group.add_argument(
            "--strip-suffix",
            action="append",
            default=[],
            help="Remove this suffix from read names if present. Can be given "
            "multiple times. (no)",
        )
        group.add_argument(
            "--length-tag",
            metavar="TAG",
            help="Search for TAG followed by a decimal number in the "
            "description field of the read. Replace the decimal number "
            "with the correct length of the trimmed read. For example, "
            "use --length-tag 'length=' to correct fields like "
            "'length=123'. (no)",
        )

        group = parser.add_group("UMIs", title="UMI options")
        group.add_argument(
            "--read1-umi",
            type=int,
            default=None,
            metavar="N",
            help="Clip N UMI bases from the 5' end of read 1 and append them "
            "to the read name. Mutually exclusive with '--umi-input'. (default: 0)",
        )
        group.add_argument(
            "--read2-umi",
            type=int,
            default=None,
            metavar="N",
            help="Clip N UMI bases from the 5' end of read 2 and append them "
            "to the read name. Mutually exclusive with '--umi-input'. (default: 0)",
        )
        group.add_argument(
            "--umi-delim",
            default=":",
            metavar="CHAR",
            help="Delimiter for separating UMI from read ID. (default: ':')",
        )

        group = parser.add_group("Filtering", title="Filtering of processed reads")
        group.add_argument(
            "--discard-trimmed",
            action="store_true",
            default=False,
            help="Discard reads that contain an adapter. Also use -O to avoid "
            "discarding too many randomly matching reads! (no)",
        )
        group.add_argument(
            "--discard-untrimmed",
            action="store_true",
            default=False,
            help="Discard reads that do not contain the adapter. (no)",
        )
        group.add_argument(
            "-m",
            "--minimum-length",
            type=positive(int, True),
            default=None,
            metavar="LENGTH",
            help="Discard trimmed reads that are shorter than LENGTH. "
            "Reads that are too short even before adapter removal are "
            "also discarded. In colorspace, an initial primer is not "
            "counted. (0)",
        )
        group.add_argument(
            "-M",
            "--maximum-length",
            type=positive(int, True),
            default=sys.maxsize,
            metavar="LENGTH",
            help="Discard trimmed reads that are longer than LENGTH. "
            "Reads that are too long even before adapter removal "
            "are also discarded. In colorspace, an initial primer "
            "is not counted. (no limit)",
        )
        group.add_argument(
            "--max-n",
            type=positive(float, True),
            default=None,
            metavar="COUNT",
            help="Discard reads with too many N bases. If COUNT is an integer, "
            "it is treated as the absolute number of N bases. If it is "
            "between 0 and 1, it is treated as the proportion of N's "
            "allowed in a read. (no)",
        )

        group = parser.add_group("Output")
        group.add_argument(
            "-o",
            "--output",
            type=writable_file,
            metavar="FILE",
            help="Write trimmed reads to FILE. Use '{name}' in FILE to demultiplex "
            "reads into multiple files. Use '-' to denote stdout. Unless writing to "
            "stdout, the summary report goes to stdout by default. (write to standard "
            "output)",
        )
        group.add_argument(
            "--output-format",
            action=EnumNameAction,
            const=SequenceFileType,
            metavar="FORMAT",
            help="The format of the output file. If not specified, the output "
            "format is determined from the filename. Defaults to FASTQ when writing to "
            "stdout. Note: BAM output is not (yet) supported, but you can output to "
            "stdout and pipe to 'samtools view'. (autodetect)",
        )
        group.add_argument(
            "-C",
            "--compression-format",
            metavar="FORMAT",
            default=None,
            choices=sorted(FORMATS.list_compression_formats()),
            help="Force output to be compressed using the specified format, otherwise "
            "the format is guessed from the file extension. This is option is required "
            "in order to have compressed output written to stdout/stderr. (None)"
        )
        group.add_argument(
            "--info-file",
            type=writable_file,
            metavar="FILE",
            help="Write information about each read and its adapter matches into FILE. "
            "See the documentation for the file format. (no)",
        )
        group.add_argument(
            "-r",
            "--rest-file",
            type=writable_file,
            metavar="FILE",
            help="When the adapter matches in the middle of a read, write the "
            "rest (after the adapter) into FILE. (no)",
        )
        group.add_argument(
            "--wildcard-file",
            type=writable_file,
            metavar="FILE",
            help="When the adapter has N bases (wildcards), write adapter "
            "bases matching wildcard positions to FILE. When there are "
            "indels in the alignment, this will often not be accurate. (no)",
        )
        group.add_argument(
            "--too-short-output",
            type=writable_file,
            metavar="FILE",
            help="Write reads that are too short (according to length "
            "specified by -m) to FILE. (no - too short reads are discarded)",
        )
        group.add_argument(
            "--too-long-output",
            type=writable_file,
            metavar="FILE",
            help="Write reads that are too long (according to length specified "
            "by -M) to FILE. (no - too long reads are discarded)",
        )
        group.add_argument(
            "--untrimmed-output",
            type=writable_file,
            default=None,
            metavar="FILE",
            help="Write reads that do not contain the adapter to FILE. (no - "
            "untrimmed reads are written to default output)",
        )
        group.add_argument(
            "--merged-output",
            type=writable_file,
            default=None,
            metavar="FILE",
            help="Write reads that have been merged to this file. (merged "
            "reads are discarded)",
        )
        group.add_argument(
            "--report-file",
            type=writable_file,
            default=None,
            metavar="FILE",
            help="Write report to file rather than stdout/stderr. (no)",
        )
        group.add_argument(
            "--report-formats",
            nargs="*",
            choices=cls.list_report_formats(),
            default=None,
            metavar="FORMAT",
            help="Report type(s) to generate. If multiple, '--report-file' "
            "is treated as a prefix and the appropriate extensions are "
            "appended. If unspecified, the format is guessed from the "
            "file extension. Supported formats are: txt (legacy text "
            "format), json, yaml, pickle. See the documentation for a "
            "full description of the structured output (json/yaml/pickle "
            "formats).",
        )
        group.add_argument(
            "--metrics",
            dest="metrics",
            nargs="*",
            default=None,
            help="Which read-level statistics to compute. Can be 'pre': only compute "
            "pre-trimming metrics; 'post': only compute post-trimming metrics; or "
            "'both'. The keyword can be followed by ':' and then additional "
            "configuration parameters. E.g. 'pre:tiles' means to also collect "
            "tile-level statistics (Illumina data only), and 'pre:tiles=<regexp>' "
            "means to use the specified regular expression to extract key portions "
            "of read names to collect the tile statistics. (none)",
        )

        group = parser.add_group("Colorspace options")
        group.add_argument(
            "-d",
            "--double-encode",
            action="store_true",
            default=False,
            help="Double-encode colors (map 0,1,2,3,4 to A,C,G,T,N). (no)",
        )
        group.add_argument(
            "-t",
            "--trim-primer",
            action="store_true",
            default=False,
            help="Trim primer base and the first color (which is the "
            "transition to the first nucleotide). (no)",
        )
        group.add_argument(
            "--strip-f3",
            action="store_true",
            default=False,
            help="Strip the _F3 suffix of read names. (no)",
        )
        group.add_argument(
            "--maq",
            "--bwa",
            action="store_true",
            default=False,
            help="MAQ- and BWA-compatible colorspace output. This enables -c, "
            "-d, -t, --strip-f3 and -y '/1'. (no)",
        )
        group.add_argument(
            "--no-zero-cap",
            dest="zero_cap",
            action="store_false",
            help="Do not change negative quality values to zero in colorspace "
            "data. By default, they are since many tools have problems "
            "with negative qualities. (no)",
        )
        group.add_argument(
            "-z",
            "--zero-cap",
            action="store_true",
            help="Change negative quality values to zero. This is enabled "
            "by default when -c/--colorspace is also enabled. Use the "
            "above option to disable it. (no)",
        )

        group = parser.add_group(
            "Paired",
            title="Paired-end options",
            description="The -A/-G/-B/-U/-I options work like their -a/-b/-g/-u/-i "
            "counterparts, but are applied to the second read in each pair.",
        )
        group.add_argument(
            "-A",
            "--adapter2",
            action="append",
            dest="adapters2",
            default=[],
            metavar="ADAPTER",
            help="3' adapter to be removed from second read in a pair. (no)",
        )
        group.add_argument(
            "-G",
            "--front2",
            action="append",
            dest="front2",
            default=[],
            metavar="ADAPTER",
            help="5' adapter to be removed from second read in a pair. (no)",
        )
        group.add_argument(
            "-B",
            "--anywhere2",
            action="append",
            dest="anywhere2",
            default=[],
            metavar="ADAPTER",
            help="5'/3 adapter to be removed from second read in a pair. (no)",
        )
        group.add_argument(
            "-U",
            "--cut2",
            type=int,
            action="append",
            dest="cut2",
            default=[],
            metavar="LENGTH",
            help="Remove LENGTH bases from second read in a pair (see --cut). (no)",
        )
        group.add_argument(
            "-I",
            "--cut-min2",
            type=int,
            action="append",
            default=[],
            metavar="LENGTH",
            help="Similar to -U, except that cutting is done AFTER adapter "
            "trimming, and only if a minimum of LENGTH bases was not "
            "already removed (see --cut-min). (no)",
        )
        group.add_argument(
            "-w",
            "--overwrite-low-quality",
            type=Delimited(data_type=positive(int, True), min_len=3, max_len=3),
            default=None,
            metavar="LOWQ,HIGHQ,WINDOW",
            help="When one read has mean quality < LOWQ and the other read has "
            "mean quality >= HIGHQ over the first WINDOW bases, overwrite "
            "the worse read with the better read.",
        )
        group.add_argument(
            "-p",
            "--paired-output",
            type=writable_file,
            metavar="FILE",
            help="Write second read in a pair to FILE. (no)",
        )
        group.add_argument(
            "-L",
            "--interleaved-output",
            type=writable_file,
            metavar="FILE",
            help="Write output to interleaved file.",
        )
        group.add_argument(
            "--pair-filter",
            action=EnumNameAction,
            const=PairFilter,
            metavar="TYPE",
            help="Which of the reads in a paired-end read have to match the "
            "filtering criterion in order for it to be filtered. (any)",
        )
        group.add_argument(
            "--untrimmed-paired-output",
            type=writable_file,
            default=None,
            metavar="FILE",
            help="Write second read in a pair to this FILE when no adapter "
            "was found in the first read. Use this option together with "
            "--untrimmed-output when trimming paired-end reads. (no - "
            "output to same file as trimmed reads)",
        )
        group.add_argument(
            "--too-short-paired-output",
            type=writable_file,
            default=None,
            metavar="FILE",
            help="Write second read in a pair to this file if pair is too "
            "short. Use together with --too-short-output. (no - too short "
            "reads are discarded)",
        )
        group.add_argument(
            "--too-long-paired-output",
            type=writable_file,
            default=None,
            metavar="FILE",
            help="Write second read in a pair to this file if pair is too "
            "long. Use together with --too-long-output. (no - too long "
            "reads are discarded)",
        )

        group = parser.add_group("Method-specific options")
        group = group.add_mutually_exclusive_group()
        group.add_argument(
            "--bisulfite",
            default=False,
            metavar="METHOD",  # TODO: convert to enum
            help="Set default option values for bisulfite-treated data. The "
            "argument specifies the type of bisulfite library (rrbs, "
            "non-directional, non-directional-rrbs, truseq, epignome, or "
            "swift) or custom parameters for trimming: "
            "'<read1>[;<read2>]' where trimming parameters for each read "
            "are: '<5' cut>,<3' cut>,<include trimmed>,<only trimmed>' "
            "where 'include trimmed' is 1 or 0 for whether or not the "
            "bases already trimmed during/prior to adapter trimming "
            "should be counted towards the total bases to be cut and "
            "'only trimmed' is 1 or 0 for whether or not only trimmed "
            "reads should be further cut. (no)",
        )
        group.add_argument(
            "--mirna",
            action="store_true",
            default=False,
            help="Set default option values for miRNA data. (no)",
        )

        group = parser.add_group("Parallel", title="Parallel (multi-core) options")
        group.add_argument(
            "-T",
            "--threads",
            type=positive(int, True),
            default=None,
            metavar="THREADS",
            help="Number of threads to use for read trimming. Set to 0 to use "
            "max available threads. (Do not use multithreading)",
        )
        group.add_argument(
            "--no-writer-process",
            action="store_false",
            dest="writer_process",
            default=True,
            help="Do not use a writer process; instead, each worker thread "
            "writes its own output to a file with a '.N' suffix. (no)",
        )
        group.add_argument(
            "--preserve-order",
            action="store_true",
            default=False,
            help="Preserve order of reads in input files (ignored if "
            "--no-writer-process is set). (no)",
        )
        group.add_argument(
            "--process-timeout",
            type=positive(int, True),
            default=DEFAULT_PROCESS_TIMEOUT,
            metavar="SECONDS",
            help=f"Number of seconds process should wait before escalating "
            f"messages to ERROR level. ({DEFAULT_PROCESS_TIMEOUT})",
        )
        group.add_argument(
            "--read-queue-size",
            type=int_or_str,
            default=None,
            metavar="SIZE",
            help="Size of queue for batches of reads to be processed. "
            "(THREADS * 100)",
        )
        group.add_argument(
            "--result-queue-size",
            type=int_or_str,
            default=None,
            metavar="SIZE",
            help="Size of queue for batches of results to be written. "
            "(THREADS * 100)",
        )
        group.add_argument(
            "--compression-mode",
            action=EnumNameAction,
            const=CompressionMode,
            metavar="MODE",
            help="Where data compression should be performed. Defaults to "
            "'writer' if system-level compression can be used and "
            "(1 < threads < 8), otherwise defaults to 'worker'.",
        )

    @staticmethod
    def _validate_trim_options(options: Namespace, parser: AtroposArgumentParser):
        paired = options.paired

        # Any of these options imply paired-end and disable legacy mode
        paired_both = bool(
            options.adapters2
            or options.front2
            or options.anywhere2
            or options.cut2
            or options.cut_min2
            or options.interleaved_output
            or options.pair_filter
            or options.overwrite_low_quality
            or options.too_short_paired_output
            or options.too_long_paired_output
        )

        # If any of the options in the "paired" group are set, we implicitly
        # expect paired-end input
        paired_implicit = paired_both or bool(
             options.paired_output or options.untrimmed_paired_output
        )

        if options.output_format is None and options.output is not None:
            options.output_format = SequenceFileType.guess_from_name(
                options.output, output=True, raise_on_failure=True
            )

        if paired or paired_implicit:
            any_output = bool(options.output or options.paired_output)

            if options.interleaved_output:
                if any_output:
                    parser.error("Cannot specify both interleaved and paired output.")

                options.use_interleaved_output = True
            elif options.output_format == SequenceFileType.SAM:
                if any_output:
                    parser.error("SAM output must be specified using the -l option")

                if options.no_writer_process:
                    logger.warning(
                        "Note: output SAM files cannot be concatenated; use "
                        "'samtools merge' instead."
                    )

                options.use_interleaved_output = True
            elif not any_output or options.output in {STDOUT, STDERR}:
                # If no output files are specified, write interleaved
                # output to stdout by default.
                options.use_interleaved_output = True
            else:
                if options.output is None:
                    parser.error(
                        "When you use -p or --paired-output, you must "
                        "also use the -o option."
                    )

                if options.paired_output is None:
                    parser.error(
                        "When paired-end trimming is enabled, a second "
                        "output file needs to be specified via -p "
                        "(--paired-output)."
                    )

                if (
                    bool(options.untrimmed_output) !=
                    bool(options.untrimmed_paired_output)
                ):
                    parser.error(
                        "When trimming paired-end reads, you must use "
                        "either none or both of the --untrimmed-output/"
                        "--untrimmed-paired-output options."
                    )

                if (
                    bool(options.too_short_output) !=
                    bool(options.too_short_paired_output)
                ):
                    parser.error(
                        "When using --too-short-output with paired-end "
                        "reads, you also need to use "
                        "--too-short-paired-output"
                    )

                if (
                    bool(options.too_long_output) !=
                    bool(options.too_long_paired_output)
                ):
                    parser.error(
                        "When using --too-long-output with paired-end "
                        "reads, you also need to use "
                        "--too-long-paired-output"
                    )

            if paired_both or options.quality_cutoff or options.trim_n:
                # Full paired-end trimming when both -p and -A/-G/-B/-U given
                # Read modifications (such as quality trimming) are applied also
                # to second read.
                paired = "both"
            else:
                # Modify first read only, keep second in sync (-p given, but not
                # -A/-G/-B/-U). This exists for backwards compatibility
                # ('legacy mode').
                paired = "first"

            options.paired = paired
        else:
            if options.read2_umi:
                parser.error(
                    "Cannot use '--read2-umi' with single-end input."
                )

            if options.untrimmed_paired_output:
                parser.error(
                    "Option --untrimmed-paired-output can only be used when "
                    "trimming paired-end reads (with option -p)."
                )

        # Send report to stderr if main output will be going to stdout
        if options.output == options.report_file:
            if options.output is None:
                options.report_file = STDERR
            else:
                parser.error("'output' and 'report_file' must be different.")
        elif options.report_file is None:
            options.report_file = (STDERR if options.output is STDOUT else STDOUT)

        # TODO: This is pretty confusing logic - need to simplify
        if options.aligner == AlignerType.ADAPTER:
            if options.indels and options.indel_cost is None:
                options.indel_cost = DEFAULT_ADAPTER_INDEL_COST

            if options.overlap is None:
                # If the user specifies a max rmp, that is used for determining the
                # minimum overlap and -O is set to 1, otherwise -O is set to the old
                # default of 3.
                if options.adapter_max_rmp is None:
                    options.overlap = DEFAULT_ADAPTER_MIN_OVERLAP
                else:
                    options.overlap = DEFAULT_ADAPTER_MIN_OVERLAP_WITH_RMP
        elif options.aligner == AlignerType.INSERT:
            if paired != "both":
                parser.error("Insert aligner only works with paired-end reads")

            # TODO: should also be checking that there is exactly one 3' adapter for
            #  each read
            # TODO: have the aligner tell us whether it can be used based on options?

            if options.indels and options.indel_cost is None:
                options.indel_cost = DEFAULT_INSERT_INDEL_COST

            if options.overlap is None:
                options.overlap = DEFAULT_INSERT_MIN_OVERLAP
                if options.adapter_max_rmp is None:
                    options.adapter_max_rmp = DEFAULT_ADAPTER_MAX_RMP

            if options.insert_match_error_rate is None:
                options.insert_match_error_rate = (
                    options.error_rate or DEFAULT_INSERT_MATCH_ERROR_RATE
                )

            if options.insert_match_adapter_error_rate is None:
                options.insert_match_adapter_error_rate = (
                    options.insert_match_error_rate
                )

        if options.merge_overlapping:
            if options.merged_output is None:
                logger.warning(
                    "--merge-output is not set; merged reads will be discarded"
                )

            if options.merge_error_rate is None:
                options.merge_error_rate = (
                    options.error_rate or DEFAULT_MERGE_ERROR_RATE
                )

        if options.mirna:
            if not (options.adapters or options.front or options.anywhere):
                options.adapters = DEFAULT_MIRNA_ADAPTERS

            if options.quality_cutoff is None:
                options.quality_cutoff = DEFAULT_MINRA_QUALITY_CUTOFF
            if options.minimum_length is None:
                options.minimum_length = DEFAULT_MIRNA_MIN_LENGTH

            if options.error_rate is None:
                options.error_rate = DEFAULT_MIRNA_ERROR_RATE
        elif options.bisulfite:
            # TODO: set default adapter sequences
            # Jury is out on whether quality trimming helps. For aligners like
            # bwameth, it actually leads to worse results.
            # if options.quality_cutoff is None:
            #     options.quality_cutoff = "20,20"
            if options.bisulfite == "swift" and paired != "both":
                parser.error("Swift trimming is only compatible with paired-end reads")

            if options.bisulfite not in (
                "rrbs",
                "non-directional",
                "truseq",
                "epignome",
                "swift",
                "non-directional-rrbs",
            ):
                def parse_bisulfite_params(arg):
                    """
                    Parses bisulfite trimming command line arguments into arguments
                    to the Modifiers.
                    """
                    try:
                        bsparts = [int(part) for part in arg.split(",")]

                        if len(bsparts) != 4:
                            raise ValueError(f"Invalid 'bisulfite' argument: {arg}")

                        if bsparts[0] <= 0 and bsparts[1] <= 0:
                            return None

                        return dict(
                            zip(
                                ("lengths", "count_trimmed", "only_trimmed"),
                                (
                                    (bsparts[0], -1 * bsparts[1]),
                                    (False, True)[bsparts[2]],
                                    (False, True)[bsparts[3]],
                                ),
                            )
                        )
                    except:
                        parser.error(
                            f"Invalid format for bisulfite parameters: {arg}"
                        )

                temp = [
                    parse_bisulfite_params(arg)
                    for arg in cast(str, options.bisulfite).split(";")
                ]

                if paired == "both" and len(temp) == 1:
                    temp = [temp[0], temp[0]]
                elif paired != "both" and len(temp) > 1:
                    parser.error("Too many bisulfite parameters for single-end reads")

                options.bisulfite = temp

        if (options.read1_umi or options.read2_umi) and options.umi_input:
            parser.error(
                "UMIs may be specified using either '--umi-input' or "
                "'--read1-umi'/'--read2-umi', not both"
            )

        if options.overwrite_low_quality:
            if not paired:
                parser.error(
                    "--overwrite-low-quality is not valid for single-end reads"
                )

            if options.overwrite_low_quality[0] > options.overwrite_low_quality[1]:
                parser.error("For --overwrite-low-quality, LOWQ must be <= HIGHQ")

        if options.quality_cutoff:
            if all(c <= 0 for c in options.quality_cutoff):
                options.quality_cutoff = None
            elif len(options.quality_cutoff) == 1:
                options.quality_cutoff = [0] + options.quality_cutoff

        if options.pair_filter is None:
            options.pair_filter = PairFilter.ANY

        if (options.discard_trimmed or options.discard_untrimmed) and (
            options.untrimmed_output is not None
        ):
            parser.error(
                "Only one of the --discard-trimmed, --discard-untrimmed "
                "and --untrimmed-output options can be used at the same time."
            )

        compression_format = options.compression_format

        if options.output is not None:
            # Guess compression format from filename
            if compression_format is None:
                compression_format = FORMATS.guess_compression_format(options.output)

            # Multiplexed output
            if "{name}" in options.output.name:
                if options.discard_trimmed:
                    parser.error("Do not use --discard-trimmed when demultiplexing.")

                if paired:
                    parser.error(
                        "Demultiplexing not yet supported for paired-end files."
                    )

        if compression_format is not None:
            fmt = FORMATS.get_compression_format(compression_format)
            options.can_use_system_compression = fmt.can_use_system_compression()

        if options.maq:
            options.colorspace = True
            options.double_encode = True
            options.trim_primer = True
            options.suffix = "/1"

        if options.strip_f3 or options.maq:
            options.strip_suffix.append("_F3")

        if options.zero_cap is None:
            options.zero_cap = options.colorspace

        if options.colorspace:
            if options.anywhere:
                parser.error(
                    "Using --anywhere with colorspace reads is currently not "
                    "supported (if you think this may be useful, contact the "
                    "author)."
                )

            if options.match_read_wildcards:
                parser.error("IUPAC wildcards not supported in colorspace")

            options.match_adapter_wildcards = False
        else:
            if options.trim_primer:
                parser.error("Trimming the primer makes only sense in colorspace.")

            if options.double_encode:
                parser.error("Double-encoding makes only sense in colorspace.")

        if options.error_rate is None:
            options.error_rate = DEFAULT_ERROR_RATE

        if options.cut:
            if len(options.cut) > 2:
                parser.error("You cannot remove bases from more than two ends.")

            if len(options.cut) == 2 and options.cut[0] * options.cut[1] > 0:
                parser.error("You cannot remove bases from the same end twice.")

        if options.cut_min:
            if len(options.cut_min) > 2:
                parser.error("You cannot remove bases from more than two ends.")

            if (
                len(options.cut_min) == 2
                and options.cut_min[0] * options.cut_min[1] > 0
            ):
                parser.error("You cannot remove bases from the same end twice.")

        if paired == "both" and options.cut2:
            if len(options.cut2) > 2:
                parser.error("You cannot remove bases from more than two ends.")

            if len(options.cut2) == 2 and options.cut2[0] * options.cut2[1] > 0:
                parser.error("You cannot remove bases from the same end twice.")

        if paired == "both" and options.cut_min2:
            if len(options.cut_min2) > 2:
                parser.error("You cannot remove bases from more than two ends.")

            if (
                len(options.cut_min2) == 2
                and options.cut_min2[0] * options.cut_min2[1] > 0
            ):
                parser.error("You cannot remove bases from the same end twice.")

        if not options.metrics or options.metrics == "none":
            options.metrics = None
        else:
            metrics = {}

            for metric_spec in options.metrics:
                parts = metric_spec.split(":")
                name = parts[0]
                args = {} if len(parts) == 1 else parse_metrics_args(parts[1])
                if name == "both":
                    metrics[MetricsMode.PRE] = metrics[MetricsMode.PRE] = args
                else:
                    metrics[MetricsMode(name)] = args

            options.metrics = metrics

        if options.threads is not None:
            threads = configure_threads(options, parser)

            if options.compression_mode is None:
                options.compression_mode = choose_compression_mode(options)
            elif options.compression_mode == CompressionMode.WRITER:
                if not options.writer_process:
                    parser.error(
                        "Writer compression and --no-writer-process are "
                        "mutually exclusive"
                    )
                elif threads == 2:
                    logger.warning(
                        "Writer compression requires > 2 threads; using worker "
                        "compression instead"
                    )
                    options.compression_mode = CompressionMode.WORKER

            # Set queue sizes if necessary.
            # If we are using writer compression, the back-up will be in the
            # result queue, otherwise it will be in the read queue.
            if options.read_queue_size is None:
                if options.compression_mode == CompressionMode.WRITER:
                    read_queue_multiplier = WRITER_READ_QUEUE_SIZE_MULTIPLER
                else:
                    read_queue_multiplier = WORKER_READ_QUEUE_SIZE_MULTIPLIER

                options.read_queue_size = threads * read_queue_multiplier
            elif 0 < options.read_queue_size < threads:
                parser.error("Read queue size must be >= 'threads'")

            if options.result_queue_size is None:
                if options.compression_mode == CompressionMode.WORKER:
                    result_queue_multiplier = WORKER_RESULT_QUEUE_SIZE_MULTIPLIER
                else:
                    result_queue_multiplier = WRITER_RESULT_QUEUE_SIZE_MULTIPLIER

                options.result_queue_size = threads * result_queue_multiplier
            elif 0 < options.result_queue_size < threads:
                parser.error("Result queue size must be >= 'threads'")

            max_queue_size = options.read_queue_size + options.result_queue_size

            if options.batch_size is None:
                options.batch_size = max(
                    DEFAULT_BATCH_SIZE, int(max_queue_size / MAX_QUEUE_SIZE)
                )
            elif options.batch_size * max_queue_size > MAX_QUEUE_SIZE:
                logger.warning(
                    "Combination of batch size {options.batch_size} and total queue "
                    "size {max_queue_size} may lead to excessive memory usage"
                )

        if options.batch_size is None:
            options.batch_size = DEFAULT_BATCH_SIZE
